#!/usr/bin/env python
#
# Copyright 2011, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Tests for memorizingfile module."""

from __future__ import absolute_import

import unittest

import six

import set_sys_path  # Update sys.path to locate pywebsocket3 module.
from pywebsocket3 import memorizingfile


class UtilTest(unittest.TestCase):
    """A unittest for memorizingfile module."""
    def check(self, memorizing_file, num_read, expected_list):
        for unused in range(num_read):
            memorizing_file.readline()
        actual_list = memorizing_file.get_memorized_lines()
        self.assertEqual(len(expected_list), len(actual_list))
        for expected, actual in zip(expected_list, actual_list):
            self.assertEqual(expected, actual)

    def check_with_size(self, memorizing_file, read_size, expected_list):
        read_list = []
        read_line = ''
        while True:
            line = memorizing_file.readline(read_size)
            line_length = len(line)
            self.assertTrue(line_length <= read_size)
            if line_length == 0:
                if read_line != '':
                    read_list.append(read_line)
                break
            read_line += line
            if line[line_length - 1] == '\n':
                read_list.append(read_line)
                read_line = ''
        actual_list = memorizing_file.get_memorized_lines()
        self.assertEqual(len(expected_list), len(actual_list))
        self.assertEqual(len(expected_list), len(read_list))
        for expected, actual, read in zip(expected_list, actual_list,
                                          read_list):
            self.assertEqual(expected, actual)
            self.assertEqual(expected, read)

    def test_get_memorized_lines(self):
        memorizing_file = memorizingfile.MemorizingFile(
            six.StringIO('Hello\nWorld\nWelcome'))
        self.check(memorizing_file, 3, ['Hello\n', 'World\n', 'Welcome'])

    def test_get_memorized_lines_limit_memorized_lines(self):
        memorizing_file = memorizingfile.MemorizingFile(
            six.StringIO('Hello\nWorld\nWelcome'), 2)
        self.check(memorizing_file, 3, ['Hello\n', 'World\n'])

    def test_get_memorized_lines_empty_file(self):
        memorizing_file = memorizingfile.MemorizingFile(six.StringIO(''))
        self.check(memorizing_file, 10, [])

    def test_get_memorized_lines_with_size(self):
        for size in range(1, 10):
            memorizing_file = memorizingfile.MemorizingFile(
                six.StringIO('Hello\nWorld\nWelcome'))
            self.check_with_size(memorizing_file, size,
                                 ['Hello\n', 'World\n', 'Welcome'])


if __name__ == '__main__':
    unittest.main()

# vi:sts=4 sw=4 et



#!/usr/bin/env python

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import collections
import json
import os

import mozhttpd
import mozunit
import pytest
from six import ensure_binary, ensure_str
from six.moves.urllib.error import HTTPError
from six.moves.urllib.request import (
    HTTPHandler,
    ProxyHandler,
    Request,
    build_opener,
    install_opener,
    urlopen,
)


def httpd_url(httpd, path, querystr=None):
    """Return the URL to a started MozHttpd server for the given info."""

    url = "http://127.0.0.1:{port}{path}".format(
        port=httpd.httpd.server_port,
        path=path,
    )

    if querystr is not None:
        url = "{url}?{querystr}".format(
            url=url,
            querystr=querystr,
        )

    return url


@pytest.fixture(name="num_requests")
def fixture_num_requests():
    """Return a defaultdict to count requests to HTTP handlers."""
    return collections.defaultdict(int)


@pytest.fixture(name="try_get")
def fixture_try_get(num_requests):
    """Return a function to try GET requests to the server."""

    def try_get(httpd, querystr):
        """Try GET requests to the server."""

        num_requests["get_handler"] = 0

        f = urlopen(httpd_url(httpd, "/api/resource/1", querystr))

        assert f.getcode() == 200
        assert json.loads(f.read()) == {"called": 1, "id": "1", "query": querystr}
        assert num_requests["get_handler"] == 1

    return try_get


@pytest.fixture(name="try_post")
def fixture_try_post(num_requests):
    """Return a function to try POST calls to the server."""

    def try_post(httpd, querystr):
        """Try POST calls to the server."""

        num_requests["post_handler"] = 0

        postdata = {"hamburgers": "1234"}

        f = urlopen(
            httpd_url(httpd, "/api/resource/", querystr),
            data=ensure_binary(json.dumps(postdata)),
        )

        assert f.getcode() == 201
        assert json.loads(f.read()) == {
            "called": 1,
            "data": postdata,
            "query": querystr,
        }
        assert num_requests["post_handler"] == 1

    return try_post


@pytest.fixture(name="try_del")
def fixture_try_del(num_requests):
    """Return a function to try DEL calls to the server."""

    def try_del(httpd, querystr):
        """Try DEL calls to the server."""

        num_requests["del_handler"] = 0

        opener = build_opener(HTTPHandler)
        request = Request(httpd_url(httpd, "/api/resource/1", querystr))
        request.get_method = lambda: "DEL"
        f = opener.open(request)

        assert f.getcode() == 200
        assert json.loads(f.read()) == {"called": 1, "id": "1", "query": querystr}
        assert num_requests["del_handler"] == 1

    return try_del


@pytest.fixture(name="httpd_no_urlhandlers")
def fixture_httpd_no_urlhandlers():
    """Yields a started MozHttpd server with no URL handlers."""
    httpd = mozhttpd.MozHttpd(port=0)
    httpd.start(block=False)
    yield httpd
    httpd.stop()


@pytest.fixture(name="httpd_with_docroot")
def fixture_httpd_with_docroot(num_requests):
    """Yields a started MozHttpd server with docroot set."""

    @mozhttpd.handlers.json_response
    def get_handler(request, objid):
        """Handler for HTTP GET requests."""

        num_requests["get_handler"] += 1

        return (
            200,
            {
                "called": num_requests["get_handler"],
                "id": objid,
                "query": request.query,
            },
        )

    httpd = mozhttpd.MozHttpd(
        port=0,
        docroot=os.path.dirname(os.path.abspath(__file__)),
        urlhandlers=[
            {
                "method": "GET",
                "path": "/api/resource/([^/]+)/?",
                "function": get_handler,
            }
        ],
    )

    httpd.start(block=False)
    yield httpd
    httpd.stop()


@pytest.fixture(name="httpd")
def fixture_httpd(num_requests):
    """Yield a started MozHttpd server."""

    @mozhttpd.handlers.json_response
    def get_handler(request, objid):
        """Handler for HTTP GET requests."""

        num_requests["get_handler"] += 1

        return (
            200,
            {
                "called": num_requests["get_handler"],
                "id": objid,
                "query": request.query,
            },
        )

    @mozhttpd.handlers.json_response
    def post_handler(request):
        """Handler for HTTP POST requests."""

        num_requests["post_handler"] += 1

        return (
            201,
            {
                "called": num_requests["post_handler"],
                "data": json.loads(request.body),
                "query": request.query,
            },
        )

    @mozhttpd.handlers.json_response
    def del_handler(request, objid):
        """Handler for HTTP DEL requests."""

        num_requests["del_handler"] += 1

        return (
            200,
            {
                "called": num_requests["del_handler"],
                "id": objid,
                "query": request.query,
            },
        )

    httpd = mozhttpd.MozHttpd(
        port=0,
        urlhandlers=[
            {
                "method": "GET",
                "path": "/api/resource/([^/]+)/?",
                "function": get_handler,
            },
            {
                "method": "POST",
                "path": "/api/resource/?",
                "function": post_handler,
            },
            {
                "method": "DEL",
                "path": "/api/resource/([^/]+)/?",
                "function": del_handler,
            },
        ],
    )

    httpd.start(block=False)
    yield httpd
    httpd.stop()


def test_api(httpd, try_get, try_post, try_del):
    # GET requests
    try_get(httpd, "")
    try_get(httpd, "?foo=bar")

    # POST requests
    try_post(httpd, "")
    try_post(httpd, "?foo=bar")

    # DEL requests
    try_del(httpd, "")
    try_del(httpd, "?foo=bar")

    # GET: By default we don't serve any files if we just define an API
    with pytest.raises(HTTPError) as exc_info:
        urlopen(httpd_url(httpd, "/"))

    assert exc_info.value.code == 404


def test_nonexistent_resources(httpd_no_urlhandlers):
    # GET: Return 404 for non-existent endpoint
    with pytest.raises(HTTPError) as excinfo:
        urlopen(httpd_url(httpd_no_urlhandlers, "/api/resource/"))
    assert excinfo.value.code == 404

    # POST: POST should also return 404
    with pytest.raises(HTTPError) as excinfo:
        urlopen(
            httpd_url(httpd_no_urlhandlers, "/api/resource/"),
            data=ensure_binary(json.dumps({})),
        )
    assert excinfo.value.code == 404

    # DEL: DEL should also return 404
    opener = build_opener(HTTPHandler)
    request = Request(httpd_url(httpd_no_urlhandlers, "/api/resource/"))
    request.get_method = lambda: "DEL"

    with pytest.raises(HTTPError) as excinfo:
        opener.open(request)
    assert excinfo.value.code == 404


def test_api_with_docroot(httpd_with_docroot, try_get):
    f = urlopen(httpd_url(httpd_with_docroot, "/"))
    assert f.getcode() == 200
    assert "Directory listing for" in ensure_str(f.read())

    # Make sure API methods still work
    try_get(httpd_with_docroot, "")
    try_get(httpd_with_docroot, "?foo=bar")


def index_contents(host):
    """Return the expected index contents for the given host."""
    return "{host} index".format(host=host)


@pytest.fixture(name="hosts")
def fixture_hosts():
    """Returns a tuple of hosts."""
    return ("mozilla.com", "mozilla.org")


@pytest.fixture(name="docroot")
def fixture_docroot(tmpdir):
    """Returns a path object to a temporary docroot directory."""
    docroot = tmpdir.mkdir("docroot")
    index_file = docroot.join("index.html")
    index_file.write(index_contents("*"))

    yield docroot

    docroot.remove()


@pytest.fixture(name="httpd_with_proxy_handler")
def fixture_httpd_with_proxy_handler(docroot):
    """Yields a started MozHttpd server for the proxy test."""

    httpd = mozhttpd.MozHttpd(port=0, docroot=str(docroot))
    httpd.start(block=False)

    port = httpd.httpd.server_port
    proxy_support = ProxyHandler(
        {
            "http": "http://127.0.0.1:{port:d}".format(port=port),
        }
    )
    install_opener(build_opener(proxy_support))

    yield httpd

    httpd.stop()

    # Reset proxy opener in case it changed
    install_opener(None)


def test_proxy(httpd_with_proxy_handler, hosts):
    for host in hosts:
        f = urlopen("http://{host}/".format(host=host))
        assert f.getcode() == 200
        assert f.read() == ensure_binary(index_contents("*"))


@pytest.fixture(name="httpd_with_proxy_host_dirs")
def fixture_httpd_with_proxy_host_dirs(docroot, hosts):
    for host in hosts:
        index_file = docroot.mkdir(host).join("index.html")
        index_file.write(index_contents(host))

    httpd = mozhttpd.MozHttpd(port=0, docroot=str(docroot), proxy_host_dirs=True)

    httpd.start(block=False)

    port = httpd.httpd.server_port
    proxy_support = ProxyHandler(
        {"http": "http://127.0.0.1:{port:d}".format(port=port)}
    )
    install_opener(build_opener(proxy_support))

    yield httpd

    httpd.stop()

    # Reset proxy opener in case it changed
    install_opener(None)


def test_proxy_separate_directories(httpd_with_proxy_host_dirs, hosts):
    for host in hosts:
        f = urlopen("http://{host}/".format(host=host))
        assert f.getcode() == 200
        assert f.read() == ensure_binary(index_contents(host))

    unproxied_host = "notmozilla.org"

    with pytest.raises(HTTPError) as excinfo:
        urlopen("http://{host}/".format(host=unproxied_host))

    assert excinfo.value.code == 404


if __name__ == "__main__":
    mozunit.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os

from setuptools import setup

VERSION = "1.0.0"
HERE = os.path.dirname(__file__)
README = open(os.path.join(HERE, "README.rst")).read()

setup(
    name="mach",
    description="Generic command line command dispatching framework.",
    long_description=README,
    license="MPL 2.0",
    author="Gregory Szorc",
    author_email="gregory.szorc@gmail.com",
    url="https://developer.mozilla.org/en-US/docs/Developer_Guide/mach",
    packages=["mach", "mach.mixin"],
    version=VERSION,
    classifiers=[
        "Environment :: Console",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Natural Language :: English",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
    ],
    install_requires=[
        "blessed",
        "mozfile",
        "mozprocess",
        "six",
    ],
    tests_require=["mock"],
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import glob
import os
import platform
import posixpath
import re
import shutil
import signal
import subprocess
import sys
import telnetlib
import time
from enum import Enum

import six
from mozdevice import ADBDeviceFactory, ADBHost
from six.moves import input, urllib

MOZBUILD_PATH = os.environ.get(
    "MOZBUILD_STATE_PATH", os.path.expanduser(os.path.join("~", ".mozbuild"))
)

EMULATOR_HOME_DIR = os.path.join(MOZBUILD_PATH, "android-device")

EMULATOR_AUTH_FILE = os.path.join(
    os.path.expanduser("~"), ".emulator_console_auth_token"
)

TOOLTOOL_PATH = "python/mozbuild/mozbuild/action/tooltool.py"

TRY_URL = "https://hg.mozilla.org/try/raw-file/default"

MANIFEST_PATH = "testing/config/tooltool-manifests"

SHORT_TIMEOUT = 10

verbose_logging = False

LLDB_SERVER_INSTALL_COMMANDS_SCRIPT = """
umask 0002

mkdir -p {lldb_bin_dir}

cp /data/local/tmp/lldb-server {lldb_bin_dir}
chmod +x {lldb_bin_dir}/lldb-server

chmod 0775 {lldb_dir}
""".lstrip()

LLDB_SERVER_START_COMMANDS_SCRIPT = """
umask 0002

export LLDB_DEBUGSERVER_LOG_FILE={lldb_log_file}
export LLDB_SERVER_LOG_CHANNELS="{lldb_log_channels}"
export LLDB_DEBUGSERVER_DOMAINSOCKET_DIR={socket_dir}

rm -rf {lldb_tmp_dir}
mkdir {lldb_tmp_dir}
export TMPDIR={lldb_tmp_dir}

rm -rf {lldb_log_dir}
mkdir {lldb_log_dir}

touch {lldb_log_file}
touch {platform_log_file}

cd {lldb_tmp_dir}
{lldb_bin_dir}/lldb-server platform --server --listen {listener_scheme}://{socket_file} \\
    --log-file "{platform_log_file}" --log-channels "{lldb_log_channels}" \\
    < /dev/null > {platform_stdout_log_file} 2>&1 &
""".lstrip()


class InstallIntent(Enum):
    YES = 1
    NO = 2


class AvdInfo(object):
    """
    Simple class to contain an AVD description.
    """

    def __init__(self, description, name, extra_args, x86):
        self.description = description
        self.name = name
        self.extra_args = extra_args
        self.x86 = x86


"""
   A dictionary to map an AVD type to a description of that type of AVD.

   There is one entry for each type of AVD used in Mozilla automated tests
   and the parameters for each reflect those used in mozharness.
"""
AVD_DICT = {
    "arm": AvdInfo(
        "Android arm",
        "mozemulator-armeabi-v7a",
        [
            "-skip-adb-auth",
            "-verbose",
            "-show-kernel",
            "-ranchu",
            "-selinux",
            "permissive",
            "-memory",
            "3072",
            "-cores",
            "4",
            "-skin",
            "800x1280",
            "-gpu",
            "on",
            "-no-snapstorage",
            "-no-snapshot",
            "-prop",
            "ro.test_harness=true",
        ],
        False,
    ),
    "arm64": AvdInfo(
        "Android arm64",
        "mozemulator-arm64",
        [
            "-skip-adb-auth",
            "-verbose",
            "-show-kernel",
            "-ranchu",
            "-selinux",
            "permissive",
            "-memory",
            "3072",
            "-cores",
            "4",
            "-skin",
            "800x1280",
            "-gpu",
            "on",
            "-no-snapstorage",
            "-no-snapshot",
            "-prop",
            "ro.test_harness=true",
        ],
        False,
    ),
    "x86_64": AvdInfo(
        "Android x86_64",
        "mozemulator-x86_64",
        [
            "-skip-adb-auth",
            "-verbose",
            "-show-kernel",
            "-ranchu",
            "-selinux",
            "permissive",
            "-memory",
            "3072",
            "-cores",
            "4",
            "-skin",
            "800x1280",
            "-prop",
            "ro.test_harness=true",
            "-no-snapstorage",
            "-no-snapshot",
        ],
        True,
    ),
}


def _get_device(substs, device_serial=None):
    adb_path = _find_sdk_exe(substs, "adb", False)
    if not adb_path:
        adb_path = "adb"
    device = ADBDeviceFactory(
        adb=adb_path, verbose=verbose_logging, device=device_serial
    )
    return device


def _install_host_utils(build_obj):
    _log_info("Installing host utilities...")
    installed = False
    host_platform = _get_host_platform()
    if host_platform:
        path = os.path.join(build_obj.topsrcdir, MANIFEST_PATH)
        path = os.path.join(path, host_platform, "hostutils.manifest")
        _get_tooltool_manifest(
            build_obj.substs, path, EMULATOR_HOME_DIR, "releng.manifest"
        )
        _tooltool_fetch(build_obj.substs)
        xre_path = glob.glob(os.path.join(EMULATOR_HOME_DIR, "host-utils*"))
        for path in xre_path:
            if os.path.isdir(path) and os.path.isfile(
                os.path.join(path, _get_xpcshell_name())
            ):
                os.environ["MOZ_HOST_BIN"] = path
                installed = True
            elif os.path.isfile(path):
                os.remove(path)
        if not installed:
            _log_warning("Unable to install host utilities.")
    else:
        _log_warning(
            "Unable to install host utilities -- your platform is not supported!"
        )


def _get_xpcshell_name():
    """
    Returns the xpcshell binary's name as a string (dependent on operating system).
    """
    xpcshell_binary = "xpcshell"
    if os.name == "nt":
        xpcshell_binary = "xpcshell.exe"
    return xpcshell_binary


def _maybe_update_host_utils(build_obj):
    """
    Compare the installed host-utils to the version name in the manifest;
    if the installed version is older, offer to update.
    """

    # Determine existing/installed version
    existing_path = None
    xre_paths = glob.glob(os.path.join(EMULATOR_HOME_DIR, "host-utils*"))
    for path in xre_paths:
        if os.path.isdir(path) and os.path.isfile(
            os.path.join(path, _get_xpcshell_name())
        ):
            existing_path = path
            break
    if existing_path is None:
        # if not installed, no need to upgrade (new version will be installed)
        return
    existing_version = os.path.basename(existing_path)

    # Determine manifest version
    manifest_version = None
    host_platform = _get_host_platform()
    if host_platform:
        # Extract tooltool file name from manifest, something like:
        #     "filename": "host-utils-58.0a1.en-US-linux-x86_64.tar.gz",
        path = os.path.join(build_obj.topsrcdir, MANIFEST_PATH)
        manifest_path = os.path.join(path, host_platform, "hostutils.manifest")
        with open(manifest_path, "r") as f:
            for line in f.readlines():
                m = re.search('.*"(host-utils-.*)"', line)
                if m:
                    manifest_version = m.group(1)
                    break

    # Compare, prompt, update
    if existing_version and manifest_version:
        hu_version_regex = r"host-utils-([\d\.]*)"
        manifest_version = float(re.search(hu_version_regex, manifest_version).group(1))
        existing_version = float(re.search(hu_version_regex, existing_version).group(1))
        if existing_version < manifest_version:
            _log_info("Your host utilities are out of date!")
            _log_info(
                "You have %s installed, but %s is available"
                % (existing_version, manifest_version)
            )
            response = input("Update host utilities? (Y/n) ").strip()
            if response.lower().startswith("y") or response == "":
                parts = os.path.split(existing_path)
                backup_dir = "_backup-" + parts[1]
                backup_path = os.path.join(parts[0], backup_dir)
                shutil.move(existing_path, backup_path)
                _install_host_utils(build_obj)


def verify_android_device(
    build_obj,
    install=InstallIntent.NO,
    xre=False,
    debugger=False,
    network=False,
    verbose=False,
    app=None,
    device_serial=None,
    aab=False,
):
    """
    Determine if any Android device is connected via adb.
    If no device is found, prompt to start an emulator.
    If a device is found or an emulator started and 'install' is
    specified, also check whether Firefox is installed on the
    device; if not, prompt to install Firefox.
    If 'xre' is specified, also check with MOZ_HOST_BIN is set
    to a valid xre/host-utils directory; if not, prompt to set
    one up.
    If 'debugger' is specified, also check that lldb-server is installed;
    if it is not found, set it up.
    If 'network' is specified, also check that the device has basic
    network connectivity.
    Returns True if the emulator was started or another device was
    already connected.
    """
    if "MOZ_DISABLE_ADB_INSTALL" in os.environ or install == InstallIntent.NO:
        install = InstallIntent.NO
        _log_info(
            "Found MOZ_DISABLE_ADB_INSTALL in environment and/or the"
            " --noinstall flag, skipping android app installation"
        )
    else:
        _log_info(
            "*********************************************************************\n"
            "Neither the MOZ_DISABLE_ADB_INSTALL environment variable nor the\n"
            "--noinstall flag was found. The code will now uninstall the current\n"
            "app then re-install the android app from a different source. If you\n"
            "don't want this set your local env so that\n"
            "MOZ_DISABLE_ADB_INSTALL=True or pass the --noinstall flag\n"
            "*********************************************************************"
        )
    device_verified = False
    emulator = AndroidEmulator("*", substs=build_obj.substs, verbose=verbose)
    adb_path = _find_sdk_exe(build_obj.substs, "adb", False)
    if not adb_path:
        adb_path = "adb"
    adbhost = ADBHost(adb=adb_path, verbose=verbose, timeout=SHORT_TIMEOUT)
    devices = adbhost.devices(timeout=SHORT_TIMEOUT)
    if "device" in [d["state"] for d in devices]:
        device_verified = True
    elif emulator.is_available():
        response = input(
            "No Android devices connected. Start an emulator? (Y/n) "
        ).strip()
        if response.lower().startswith("y") or response == "":
            if not emulator.check_avd():
                _log_info("Android AVD not found, please run |mach bootstrap|")
                return
            _log_info(
                "Starting emulator running %s..." % emulator.get_avd_description()
            )
            emulator.start()
            emulator.wait_for_start()
            device_verified = True

    if device_verified and "DEVICE_SERIAL" not in os.environ:
        devices = adbhost.devices(timeout=SHORT_TIMEOUT)
        for d in devices:
            if d["state"] == "device":
                os.environ["DEVICE_SERIAL"] = d["device_serial"]
                break

    if device_verified and install != InstallIntent.NO:
        # Determine if test app is installed on the device; if not,
        # prompt to install. This feature allows a test command to
        # launch an emulator, install the test app, and proceed with testing
        # in one operation. It is also a basic safeguard against other
        # cases where testing is requested but test app installation has
        # been forgotten.
        # If a test app is installed, there is no way to determine whether
        # the current build is installed, and certainly no way to
        # determine if the installed build is the desired build.
        # Installing every time (without prompting) is problematic because:
        #  - it prevents testing against other builds (downloaded apk)
        #  - installation may take a couple of minutes.
        if not app:
            app = "org.mozilla.geckoview.test_runner"
        device = _get_device(build_obj.substs, device_serial)
        response = ""
        installed = device.is_app_installed(app)

        if not installed:
            _log_info("It looks like %s is not installed on this device." % app)
        if "fennec" in app or "firefox" in app:
            if installed:
                device.uninstall_app(app)
            _log_info("Installing Firefox...")
            build_obj._run_make(directory=".", target="install", ensure_exit_code=False)
        elif app == "org.mozilla.geckoview.test":
            if installed:
                device.uninstall_app(app)
            _log_info("Installing geckoview AndroidTest...")
            build_obj._mach_context.commands.dispatch(
                "android",
                build_obj._mach_context,
                subcommand="install-geckoview-test",
                args=[],
            )
        elif app == "org.mozilla.geckoview.test_runner":
            if installed:
                device.uninstall_app(app)
            _log_info("Installing geckoview test_runner...")
            sub = (
                "install-geckoview-test_runner-aab"
                if aab
                else "install-geckoview-test_runner"
            )
            build_obj._mach_context.commands.dispatch(
                "android", build_obj._mach_context, subcommand=sub, args=[]
            )
        elif app == "org.mozilla.geckoview_example":
            if installed:
                device.uninstall_app(app)
            _log_info("Installing geckoview_example...")
            sub = (
                "install-geckoview_example-aab" if aab else "install-geckoview_example"
            )
            build_obj._mach_context.commands.dispatch(
                "android", build_obj._mach_context, subcommand=sub, args=[]
            )
        elif not installed:
            response = input(
                "It looks like %s is not installed on this device,\n"
                "but I don't know how to install it.\n"
                "Install it now, then hit Enter " % app
            )

        device.run_as_package = app

    if device_verified and xre:
        # Check whether MOZ_HOST_BIN has been set to a valid xre; if not,
        # prompt to install one.
        xre_path = os.environ.get("MOZ_HOST_BIN")
        err = None
        if not xre_path:
            err = (
                "environment variable MOZ_HOST_BIN is not set to a directory "
                "containing host xpcshell"
            )
        elif not os.path.isdir(xre_path):
            err = "$MOZ_HOST_BIN does not specify a directory"
        elif not os.path.isfile(os.path.join(xre_path, _get_xpcshell_name())):
            err = "$MOZ_HOST_BIN/xpcshell does not exist"
        if err:
            _maybe_update_host_utils(build_obj)
            xre_path = glob.glob(os.path.join(EMULATOR_HOME_DIR, "host-utils*"))
            for path in xre_path:
                if os.path.isdir(path) and os.path.isfile(
                    os.path.join(path, _get_xpcshell_name())
                ):
                    os.environ["MOZ_HOST_BIN"] = path
                    err = None
                    break
        if err:
            _log_info("Host utilities not found: %s" % err)
            response = input("Download and setup your host utilities? (Y/n) ").strip()
            if response.lower().startswith("y") or response == "":
                _install_host_utils(build_obj)

    if device_verified and network:
        # Optionally check the network: If on a device that does not look like
        # an emulator, verify that the device IP address can be obtained
        # and check that this host can ping the device.
        serial = device_serial or os.environ.get("DEVICE_SERIAL")
        if not serial or ("emulator" not in serial):
            device = _get_device(build_obj.substs, serial)
            device.run_as_package = app
            try:
                addr = device.get_ip_address()
                if not addr:
                    _log_warning("unable to get Android device's IP address!")
                    _log_warning(
                        "tests may fail without network connectivity to the device!"
                    )
                else:
                    _log_info("Android device's IP address: %s" % addr)
                    response = subprocess.check_output(["ping", "-c", "1", addr])
                    _log_debug(response)
            except Exception as e:
                _log_warning(
                    "unable to verify network connection to device: %s" % str(e)
                )
                _log_warning(
                    "tests may fail without network connectivity to the device!"
                )
        else:
            _log_debug("network check skipped on emulator")

    if debugger:
        _setup_or_run_lldb_server(app, build_obj.substs, device_serial, setup=True)

    return device_verified


def run_lldb_server(app, substs, device_serial):
    return _setup_or_run_lldb_server(app, substs, device_serial, setup=False)


def _setup_or_run_lldb_server(app, substs, device_serial, setup=True):
    device = _get_device(substs, device_serial)

    # Don't use enable_run_as here, as this will not give you what you
    # want if we have root access on the device.
    pkg_dir = device.shell_output("run-as %s pwd" % app)
    if not pkg_dir or pkg_dir == "/":
        pkg_dir = "/data/data/%s" % app
        _log_warning(
            "Unable to resolve data directory for package %s, falling back to hardcoded path"
            % app
        )

    pkg_lldb_dir = posixpath.join(pkg_dir, "lldb")
    pkg_lldb_bin_dir = posixpath.join(pkg_lldb_dir, "bin")
    pkg_lldb_server = posixpath.join(pkg_lldb_bin_dir, "lldb-server")

    if setup:
        # Check whether lldb-server is already there
        if device.shell_bool("test -x %s" % pkg_lldb_server, enable_run_as=True):
            _log_info(
                "Found lldb-server binary, terminating any running server processes..."
            )
            # lldb-server is already present. Kill any running server.
            device.shell("pkill -f lldb-server", enable_run_as=True)
        else:
            _log_info("lldb-server not found, installing...")

            # We need to do an install
            try:
                server_path_local = substs["ANDROID_LLDB_SERVER"]
            except KeyError:
                _log_info(
                    "ANDROID_LLDB_SERVER is not configured correctly; "
                    "please re-configure your build."
                )
                return

            device.push(server_path_local, "/data/local/tmp")

            install_cmds = LLDB_SERVER_INSTALL_COMMANDS_SCRIPT.format(
                lldb_bin_dir=pkg_lldb_bin_dir, lldb_dir=pkg_lldb_dir
            )

            install_cmds = [l for l in install_cmds.splitlines() if l]

            _log_debug(
                "Running the following installation commands:\n%r" % (install_cmds,)
            )

            device.batch_execute(install_cmds, enable_run_as=True)
        return

    pkg_lldb_sock_file = posixpath.join(pkg_dir, "platform-%d.sock" % int(time.time()))

    pkg_lldb_log_dir = posixpath.join(pkg_lldb_dir, "log")
    pkg_lldb_tmp_dir = posixpath.join(pkg_lldb_dir, "tmp")

    pkg_lldb_log_file = posixpath.join(pkg_lldb_log_dir, "lldb-server.log")
    pkg_platform_log_file = posixpath.join(pkg_lldb_log_dir, "platform.log")
    pkg_platform_stdout_log_file = posixpath.join(
        pkg_lldb_log_dir, "platform-stdout.log"
    )

    listener_scheme = "unix-abstract"
    log_channels = "lldb process:gdb-remote packets"

    start_cmds = LLDB_SERVER_START_COMMANDS_SCRIPT.format(
        lldb_bin_dir=pkg_lldb_bin_dir,
        lldb_log_file=pkg_lldb_log_file,
        lldb_log_channels=log_channels,
        socket_dir=pkg_dir,
        lldb_tmp_dir=pkg_lldb_tmp_dir,
        lldb_log_dir=pkg_lldb_log_dir,
        platform_log_file=pkg_platform_log_file,
        listener_scheme=listener_scheme,
        platform_stdout_log_file=pkg_platform_stdout_log_file,
        socket_file=pkg_lldb_sock_file,
    )

    start_cmds = [l for l in start_cmds.splitlines() if l]

    _log_debug("Running the following start commands:\n%r" % (start_cmds,))

    device.batch_execute(start_cmds, enable_run_as=True)

    return pkg_lldb_sock_file


def get_adb_path(build_obj):
    return _find_sdk_exe(build_obj.substs, "adb", False)


def grant_runtime_permissions(build_obj, app, device_serial=None):
    """
    Grant required runtime permissions to the specified app
    (eg. org.mozilla.geckoview.test_runner).
    """
    device = _get_device(build_obj.substs, device_serial)
    device.run_as_package = app
    device.grant_runtime_permissions(app)


class AndroidEmulator(object):
    """
    Support running the Android emulator with an AVD from Mozilla
    test automation.

    Example usage:
        emulator = AndroidEmulator()
        if not emulator.is_running() and emulator.is_available():
            if not emulator.check_avd():
                print("Android Emulator AVD not found, please run |mach bootstrap|")
            emulator.start()
            emulator.wait_for_start()
            emulator.wait()
    """

    def __init__(self, avd_type=None, verbose=False, substs=None, device_serial=None):
        global verbose_logging
        self.emulator_log = None
        self.emulator_path = "emulator"
        verbose_logging = verbose
        self.substs = substs
        self.avd_type = self._get_avd_type(avd_type)
        self.avd_info = AVD_DICT[self.avd_type]
        self.gpu = True
        self.restarted = False
        self.device_serial = device_serial
        self.avd_path = os.path.join(
            EMULATOR_HOME_DIR, "avd", "%s.avd" % self.avd_info.name
        )
        _log_debug("Running on %s" % platform.platform())
        _log_debug("Emulator created with type %s" % self.avd_type)

    def __del__(self):
        if self.emulator_log:
            self.emulator_log.close()

    def is_running(self):
        """
        Returns True if the Android emulator is running.
        """
        adb_path = _find_sdk_exe(self.substs, "adb", False)
        if not adb_path:
            adb_path = "adb"
        adbhost = ADBHost(adb=adb_path, verbose=verbose_logging, timeout=SHORT_TIMEOUT)
        devs = adbhost.devices(timeout=SHORT_TIMEOUT)
        devs = [(d["device_serial"], d["state"]) for d in devs]
        _log_debug(devs)
        if ("emulator-5554", "device") in devs:
            return True
        return False

    def is_available(self):
        """
        Returns True if an emulator executable is found.
        """
        found = False
        emulator_path = _find_sdk_exe(self.substs, "emulator", True)
        if emulator_path:
            self.emulator_path = emulator_path
            found = True
        return found

    def check_avd(self):
        """
        Determine if the AVD is already installed locally.

        Returns True if the AVD is installed.
        """
        if os.path.exists(self.avd_path):
            _log_debug("AVD found at %s" % self.avd_path)
            return True
        _log_warning("Could not find AVD at %s" % self.avd_path)
        return False

    def start(self, gpu_arg=None):
        """
        Launch the emulator.
        """
        if self.avd_info.x86 and "linux" in _get_host_platform():
            _verify_kvm(self.substs)
        if os.path.exists(EMULATOR_AUTH_FILE):
            os.remove(EMULATOR_AUTH_FILE)
            _log_debug("deleted %s" % EMULATOR_AUTH_FILE)
        self._update_avd_paths()
        # create an empty auth file to disable emulator authentication
        auth_file = open(EMULATOR_AUTH_FILE, "w")
        auth_file.close()

        env = os.environ
        env["ANDROID_EMULATOR_HOME"] = EMULATOR_HOME_DIR
        env["ANDROID_AVD_HOME"] = os.path.join(EMULATOR_HOME_DIR, "avd")
        command = [self.emulator_path, "-avd", self.avd_info.name]
        override = os.environ.get("MOZ_EMULATOR_COMMAND_ARGS")
        if override:
            command += override.split()
            _log_debug("Found MOZ_EMULATOR_COMMAND_ARGS in env: %s" % override)
        else:
            if gpu_arg:
                command += ["-gpu", gpu_arg]
                # Clear self.gpu to avoid our restart-without-gpu feature: if a specific
                # gpu setting is requested, try to use that, and nothing else.
                self.gpu = False
            elif self.gpu:
                command += ["-gpu", "on"]
            if self.avd_info.extra_args:
                # -enable-kvm option is not valid on OSX and Windows
                if (
                    _get_host_platform() in ("macosx64", "win32")
                    and "-enable-kvm" in self.avd_info.extra_args
                ):
                    self.avd_info.extra_args.remove("-enable-kvm")
                command += self.avd_info.extra_args
        log_path = os.path.join(EMULATOR_HOME_DIR, "emulator.log")
        self.emulator_log = open(log_path, "w+")
        _log_debug("Starting the emulator with this command: %s" % " ".join(command))
        _log_debug("Emulator output will be written to '%s'" % log_path)
        self.proc = subprocess.Popen(
            command,
            env=env,
            stdin=subprocess.PIPE,
            stdout=self.emulator_log,
            stderr=self.emulator_log,
        )
        _log_debug("Emulator started with pid %d" % int(self.proc.pid))

    def wait_for_start(self):
        """
        Verify that the emulator is running, the emulator device is visible
        to adb, and Android has booted.
        """
        if not self.proc:
            _log_warning("Emulator not started!")
            return False
        if self.check_completed():
            return False
        _log_debug("Waiting for device status...")
        adb_path = _find_sdk_exe(self.substs, "adb", False)
        if not adb_path:
            adb_path = "adb"
        adbhost = ADBHost(adb=adb_path, verbose=verbose_logging, timeout=SHORT_TIMEOUT)
        devs = adbhost.devices(timeout=SHORT_TIMEOUT)
        devs = [(d["device_serial"], d["state"]) for d in devs]
        while ("emulator-5554", "device") not in devs:
            time.sleep(10)
            if self.check_completed():
                return False
            devs = adbhost.devices(timeout=SHORT_TIMEOUT)
            devs = [(d["device_serial"], d["state"]) for d in devs]
        _log_debug("Device status verified.")

        _log_debug("Checking that Android has booted...")
        device = _get_device(self.substs, self.device_serial)
        complete = False
        while not complete:
            output = ""
            try:
                output = device.get_prop("sys.boot_completed", timeout=5)
            except Exception:
                # adb not yet responding...keep trying
                pass
            if output.strip() == "1":
                complete = True
            else:
                time.sleep(10)
                if self.check_completed():
                    return False
        _log_debug("Android boot status verified.")

        if not self._verify_emulator():
            return False
        if self.avd_info.x86:
            _log_info(
                "Running the x86/x86_64 emulator; be sure to install an x86 or x86_64 APK!"
            )
        else:
            _log_info("Running the arm emulator; be sure to install an arm APK!")
        return True

    def check_completed(self):
        if self.proc.poll() is not None:
            if self.gpu:
                try:
                    for line in self.emulator_log.readlines():
                        if (
                            "Invalid value for -gpu" in line
                            or "Invalid GPU mode" in line
                        ):
                            self.gpu = False
                            break
                except Exception as e:
                    _log_warning(str(e))

            if not self.gpu and not self.restarted:
                _log_warning(
                    "Emulator failed to start. Your emulator may be out of date."
                )
                _log_warning("Trying to restart the emulator without -gpu argument.")
                self.restarted = True
                self.start()
                return False
            _log_warning("Emulator has already completed!")
            log_path = os.path.join(EMULATOR_HOME_DIR, "emulator.log")
            _log_warning(
                "See log at %s and/or use --verbose for more information." % log_path
            )
            return True
        return False

    def wait(self):
        """
        Wait for the emulator to close. If interrupted, close the emulator.
        """
        try:
            self.proc.wait()
        except Exception:
            if self.proc.poll() is None:
                self.cleanup()
        return self.proc.poll()

    def cleanup(self):
        """
        Close the emulator.
        """
        self.proc.kill(signal.SIGTERM)

    def get_avd_description(self):
        """
        Return the human-friendly description of this AVD.
        """
        return self.avd_info.description

    def _update_avd_paths(self):
        ini_path = os.path.join(EMULATOR_HOME_DIR, "avd", "%s.ini" % self.avd_info.name)
        with open(ini_path, "r") as f:
            lines = f.readlines()
        with open(ini_path, "w") as f:
            for line in lines:
                if line.startswith("path="):
                    f.write("path=%s\n" % self.avd_path)
                elif line.startswith("path.rel="):
                    f.write("path.rel=avd/%s.avd\n" % self.avd_info.name)
                else:
                    f.write(line)

    def _telnet_read_until(self, telnet, expected, timeout):
        if six.PY3 and isinstance(expected, six.text_type):
            expected = expected.encode("ascii")
        return telnet.read_until(expected, timeout)

    def _telnet_write(self, telnet, command):
        if six.PY3 and isinstance(command, six.text_type):
            command = command.encode("ascii")
        telnet.write(command)

    def _telnet_cmd(self, telnet, command):
        _log_debug(">>> %s" % command)
        self._telnet_write(telnet, "%s\n" % command)
        result = self._telnet_read_until(telnet, "OK", 10)
        _log_debug("<<< %s" % result)
        return result

    def _verify_emulator(self):
        telnet_ok = False
        tn = None
        while not telnet_ok:
            try:
                tn = telnetlib.Telnet("localhost", 5554, 10)
                if tn is not None:
                    self._telnet_read_until(tn, "OK", 10)
                    self._telnet_cmd(tn, "avd status")
                    self._telnet_cmd(tn, "redir list")
                    self._telnet_cmd(tn, "network status")
                    self._telnet_write(tn, "quit\n")
                    tn.read_all()
                    telnet_ok = True
                else:
                    _log_warning("Unable to connect to port 5554")
            except Exception:
                _log_warning("Trying again after unexpected exception")
            finally:
                if tn is not None:
                    tn.close()
            if not telnet_ok:
                time.sleep(10)
                if self.proc.poll() is not None:
                    _log_warning("Emulator has already completed!")
                    return False
        return telnet_ok

    def _get_avd_type(self, requested):
        if requested in AVD_DICT.keys():
            return requested
        if self.substs:
            target_cpu = self.substs["TARGET_CPU"]
            if target_cpu == "aarch64":
                return "arm64"
            elif target_cpu.startswith("arm"):
                return "arm"
        return "x86_64"


def _find_sdk_exe(substs, exe, tools):
    if tools:
        subdirs = ["emulator", "tools"]
    else:
        subdirs = ["platform-tools"]

    found = False
    if not found and substs:
        # It's best to use the tool specified by the build, rather
        # than something we find on the PATH or crawl for.
        try:
            exe_path = substs[exe.upper()]
            if os.path.exists(exe_path):
                found = True
            else:
                _log_debug("Unable to find executable at %s" % exe_path)
        except KeyError:
            _log_debug("%s not set" % exe.upper())

    # Append '.exe' to the name on Windows if it's not present,
    # so that the executable can be found.
    if os.name == "nt" and not exe.lower().endswith(".exe"):
        exe += ".exe"

    if not found:
        # Can exe be found in the Android SDK?
        try:
            android_sdk_root = os.environ["ANDROID_SDK_ROOT"]
            for subdir in subdirs:
                exe_path = os.path.join(android_sdk_root, subdir, exe)
                if os.path.exists(exe_path):
                    found = True
                    break
                else:
                    _log_debug("Unable to find executable at %s" % exe_path)
        except KeyError:
            _log_debug("ANDROID_SDK_ROOT not set")

    if not found:
        # Can exe be found in the default bootstrap location?
        for subdir in subdirs:
            exe_path = os.path.join(MOZBUILD_PATH, "android-sdk-linux", subdir, exe)
            if os.path.exists(exe_path):
                found = True
                break
            else:
                _log_debug("Unable to find executable at %s" % exe_path)

    if not found:
        # Is exe on PATH?
        exe_path = shutil.which(exe)
        if exe_path:
            found = True
        else:
            _log_debug("Unable to find executable on PATH")

    if found:
        _log_debug("%s found at %s" % (exe, exe_path))
        try:
            creation_time = os.path.getctime(exe_path)
            _log_debug("  ...with creation time %s" % time.ctime(creation_time))
        except Exception:
            _log_warning("Could not get creation time for %s" % exe_path)

        prop_path = os.path.join(os.path.dirname(exe_path), "source.properties")
        if os.path.exists(prop_path):
            with open(prop_path, "r") as f:
                for line in f.readlines():
                    if line.startswith("Pkg.Revision"):
                        line = line.strip()
                        _log_debug(
                            "  ...with SDK version in %s: %s" % (prop_path, line)
                        )
                        break
    else:
        exe_path = None
    return exe_path


def _log_debug(text):
    if verbose_logging:
        print("DEBUG: %s" % text)


def _log_warning(text):
    print("WARNING: %s" % text)


def _log_info(text):
    print("%s" % text)


def _download_file(url, filename, path):
    _log_debug("Download %s to %s/%s..." % (url, path, filename))
    f = urllib.request.urlopen(url)
    if not os.path.isdir(path):
        try:
            os.makedirs(path)
        except Exception as e:
            _log_warning(str(e))
            return False
    local_file = open(os.path.join(path, filename), "wb")
    local_file.write(f.read())
    local_file.close()
    _log_debug("Downloaded %s to %s/%s" % (url, path, filename))
    return True


def _get_tooltool_manifest(substs, src_path, dst_path, filename):
    if not os.path.isdir(dst_path):
        try:
            os.makedirs(dst_path)
        except Exception as e:
            _log_warning(str(e))
    copied = False
    if substs and "top_srcdir" in substs:
        src = os.path.join(substs["top_srcdir"], src_path)
        if os.path.exists(src):
            dst = os.path.join(dst_path, filename)
            shutil.copy(src, dst)
            copied = True
            _log_debug("Copied tooltool manifest %s to %s" % (src, dst))
    if not copied:
        url = os.path.join(TRY_URL, src_path)
        _download_file(url, filename, dst_path)


def _tooltool_fetch(substs):
    tooltool_full_path = os.path.join(substs["top_srcdir"], TOOLTOOL_PATH)
    command = [
        sys.executable,
        tooltool_full_path,
        "fetch",
        "-o",
        "-m",
        "releng.manifest",
    ]
    try:
        response = subprocess.check_output(command, cwd=EMULATOR_HOME_DIR)
        _log_debug(response)
    except Exception as e:
        _log_warning(str(e))


def _get_host_platform():
    plat = None
    if "darwin" in str(sys.platform).lower():
        plat = "macosx64"
    elif "win32" in str(sys.platform).lower():
        plat = "win32"
    elif "linux" in str(sys.platform).lower():
        if "64" in platform.architecture()[0]:
            plat = "linux64"
        else:
            plat = "linux32"
    return plat


def _verify_kvm(substs):
    # 'emulator -accel-check' should produce output like:
    # accel:
    # 0
    # KVM (version 12) is installed and usable
    # accel
    emulator_path = _find_sdk_exe(substs, "emulator", True)
    if not emulator_path:
        emulator_path = "emulator"
    command = [emulator_path, "-accel-check"]
    try:
        out = subprocess.check_output(command)
        if six.PY3 and not isinstance(out, six.text_type):
            out = out.decode("utf-8")
        if "is installed and usable" in "".join(out):
            return
    except Exception as e:
        _log_warning(str(e))
    _log_warning("Unable to verify kvm acceleration!")
    _log_warning("The x86/x86_64 emulator may fail to start without kvm.")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


class EmulatorScreen(object):
    """Class for screen related emulator commands."""

    SO_PORTRAIT_PRIMARY = "portrait-primary"
    SO_PORTRAIT_SECONDARY = "portrait-secondary"
    SO_LANDSCAPE_PRIMARY = "landscape-primary"
    SO_LANDSCAPE_SECONDARY = "landscape-secondary"

    def __init__(self, emulator):
        self.emulator = emulator

    def initialize(self):
        self.orientation = self.SO_PORTRAIT_PRIMARY

    def _get_raw_orientation(self):
        """Get the raw value of the current device orientation."""
        response = self.emulator._run_telnet("sensor get orientation")

        return response[0].split("=")[1].strip()

    def _set_raw_orientation(self, data):
        """Set the raw value of the specified device orientation."""
        self.emulator._run_telnet("sensor set orientation %s" % data)

    def get_orientation(self):
        """Get the current device orientation.

        Returns;
            orientation -- Orientation of the device. One of:
                            SO_PORTRAIT_PRIMARY - system buttons at the bottom
                            SO_PORTRIAT_SECONDARY - system buttons at the top
                            SO_LANDSCAPE_PRIMARY - system buttons at the right
                            SO_LANDSCAPE_SECONDARY - system buttons at the left

        """
        data = self._get_raw_orientation()

        if data == "0:-90:0":
            orientation = self.SO_PORTRAIT_PRIMARY
        elif data == "0:90:0":
            orientation = self.SO_PORTRAIT_SECONDARY
        elif data == "0:0:90":
            orientation = self.SO_LANDSCAPE_PRIMARY
        elif data == "0:0:-90":
            orientation = self.SO_LANDSCAPE_SECONDARY
        else:
            raise ValueError("Unknown orientation sensor value: %s." % data)

        return orientation

    def set_orientation(self, orientation):
        """Set the specified device orientation.

        Args
            orientation -- Orientation of the device. One of:
                            SO_PORTRAIT_PRIMARY - system buttons at the bottom
                            SO_PORTRIAT_SECONDARY - system buttons at the top
                            SO_LANDSCAPE_PRIMARY - system buttons at the right
                            SO_LANDSCAPE_SECONDARY - system buttons at the left
        """
        orientation = SCREEN_ORIENTATIONS[orientation]

        if orientation == self.SO_PORTRAIT_PRIMARY:
            data = "0:-90:0"
        elif orientation == self.SO_PORTRAIT_SECONDARY:
            data = "0:90:0"
        elif orientation == self.SO_LANDSCAPE_PRIMARY:
            data = "0:0:90"
        elif orientation == self.SO_LANDSCAPE_SECONDARY:
            data = "0:0:-90"
        else:
            raise ValueError("Invalid orientation: %s" % orientation)

        self._set_raw_orientation(data)

    orientation = property(get_orientation, set_orientation)


SCREEN_ORIENTATIONS = {
    "portrait": EmulatorScreen.SO_PORTRAIT_PRIMARY,
    "landscape": EmulatorScreen.SO_LANDSCAPE_PRIMARY,
    "portrait-primary": EmulatorScreen.SO_PORTRAIT_PRIMARY,
    "landscape-primary": EmulatorScreen.SO_LANDSCAPE_PRIMARY,
    "portrait-secondary": EmulatorScreen.SO_PORTRAIT_SECONDARY,
    "landscape-secondary": EmulatorScreen.SO_LANDSCAPE_SECONDARY,
}



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os

from marionette_driver import Wait
from marionette_harness import MarionetteTestCase

"""
 Currently we expect our size after cleanup to be 98k for the current
 database schema and page sizes and constants used in this test. We
 set the threshold at 128k so the test doesn't start failing if these
 control objects/structures increase somewhat in size, but should still
 fail if we fail to delete all of the 5,000 1k files we create on disk
 as part of the test.
"""
QM_TESTING_PREF = "dom.quotaManager.testing"

EXPECTED_CACHEDIR_SIZE_AFTER_CLEANUP = 128 * 1024  # 128KB
CACHE_ID = "data"


class CachesDeleteCleanupAtShutdownTestCase(MarionetteTestCase):

    """
    Bug1784700: This test ensures that cache body files get cleaned up
    properly after cache has been deleted. Note that body files gets
    cleaned up asynchronously which means that these files might still
    be around even after Caches.Delete promise gets resolved. Currently,
    we would only clean up body files on origin initialization and that's
    a firefox is necessary here.
    """

    def setUp(self):
        super(CachesDeleteCleanupAtShutdownTestCase, self).setUp()
        self.marionette.restart(in_app=False, clean=True)
        self.marionette.set_pref(QM_TESTING_PREF, True)

    def tearDown(self):
        self.marionette.restart(in_app=False, clean=True)
        super(CachesDeleteCleanupAtShutdownTestCase, self).tearDown()
        self.marionette.set_pref(QM_TESTING_PREF, False)

    def getUsage(self):
        return self.marionette.execute_script(
            """
                return window.wrappedJSObject.getStorageEstimate();
            """,
            new_sandbox=False,
        )

    def doCacheWork(self, n):
        # max timeout for this script to execute is 5 minutes
        maxTimeout = 5 * 60 * 1000

        return self.marionette.execute_script(
            """
                const [cacheId, n] = arguments;
                return window.wrappedJSObject.doCacheWork(cacheId, n);
            """,
            script_args=(
                CACHE_ID,
                n,
            ),
            new_sandbox=False,
            script_timeout=maxTimeout,
        )

    def openCache(self):
        return self.marionette.execute_async_script(
            """
                const [cacheId, resolve] = arguments;
                window.wrappedJSObject.openCache(cacheId).then(resolve("success"));
            """,
            new_sandbox=False,
            script_args=(CACHE_ID,),
        )

    def countBodies(self):
        profile = self.marionette.instance.profile.profile
        originDir = (
            self.marionette.absolute_url("")[:-1].replace(":", "+").replace("/", "+")
        )
        morgueDir = f"{profile}/storage/default/{originDir}/cache/morgue"
        print("morgueDir path = ", morgueDir)

        bodyCount = -1
        if os.path.exists(morgueDir):
            bodyCount = 0
            for elem in os.listdir(morgueDir):
                absPathElem = os.path.join(morgueDir, elem)
                if os.path.isdir(absPathElem):
                    bodyCount += sum(
                        1
                        for e in os.listdir(absPathElem)
                        if os.path.isfile(os.path.join(absPathElem, e))
                    )
        return bodyCount

    def ensureCleanDirectory(self):
        orphanedBodiesCount = self.countBodies()
        return orphanedBodiesCount <= 0

    def isStorageInitialized(self, temporary=False):
        with self.marionette.using_context("chrome"):
            return self.marionette.execute_async_script(
                """
                    const [resolve] = arguments;
                    const req = %s;
                    req.callback = () => {
                        resolve(req.resultCode == Cr.NS_OK && req.result)
                    };
                """
                % (
                    "Services.qms.temporaryStorageInitialized()"
                    if temporary
                    else "Services.qms.storageInitialized()"
                ),
                new_sandbox=False,
            )

    def create_and_cleanup_cache(self, ensureCleanCallback, in_app):
        # create 640 cache entries
        self.doCacheWork(640)
        print("usage after doCacheWork = ", self.getUsage())

        self.marionette.restart(in_app=in_app)
        print("restart successful")

        self.marionette.navigate(
            self.marionette.absolute_url("dom/cache/cacheUsage.html")
        )
        return ensureCleanCallback()

    def afterCleanupClosure(self, usage):
        print(
            f"Storage initialized = {self.isStorageInitialized()}, temporary storage initialized = {self.isStorageInitialized(True)}"
        )

        print(f"Usage = {usage} and number of orphaned bodies = {self.countBodies()}")
        return usage < EXPECTED_CACHEDIR_SIZE_AFTER_CLEANUP

    def test_ensure_cache_cleanup_after_clean_restart(self):
        self.marionette.navigate(
            self.marionette.absolute_url("dom/cache/cacheUsage.html")
        )
        beforeUsage = self.getUsage()

        def ensureCleanCallback():
            Wait(self.marionette, interval=1, timeout=60).until(
                lambda _: self.afterCleanupClosure(self.getUsage() - beforeUsage),
                message="Cache directory is not cleaned up properly",
            )

            return (
                abs(beforeUsage - self.getUsage())
                <= EXPECTED_CACHEDIR_SIZE_AFTER_CLEANUP
                and self.ensureCleanDirectory()
            )

        if not self.create_and_cleanup_cache(ensureCleanCallback, True):
            print(f"beforeUsage = {beforeUsage}, and afterUsage = {self.getUsage()}")
            assert False

    def test_ensure_cache_cleanup_after_unclean_restart(self):
        self.marionette.navigate(
            self.marionette.absolute_url("dom/cache/cacheUsage.html")
        )

        print(f"profile path = {self.marionette.instance.profile.profile}")

        beforeUsage = self.getUsage()

        def ensureCleanCallback():
            print(
                f"ensureCleanCallback, profile path = {self.marionette.instance.profile.profile}"
            )

            self.openCache()
            Wait(self.marionette, interval=1, timeout=60).until(
                lambda _: self.afterCleanupClosure(self.getUsage() - beforeUsage),
                message="Cache directory is not cleaned up properly",
            )

            return (
                abs(beforeUsage - self.getUsage())
                <= EXPECTED_CACHEDIR_SIZE_AFTER_CLEANUP
                and self.ensureCleanDirectory()
            )

        if not self.create_and_cleanup_cache(ensureCleanCallback, False):
            print(f"beforeUsage = {beforeUsage}, and afterUsage = {self.getUsage()}")
            assert False



"""
iniconfig: brain-dead simple config-ini parsing.

compatible CPython 2.3 through to CPython 3.2, Jython, PyPy

(c) 2010 Ronny Pfannschmidt, Holger Krekel
"""

from setuptools import setup


def main():
    with open('README.txt') as fp:
        readme = fp.read()
    setup(
        name='iniconfig',
        packages=['iniconfig'],
        package_dir={'': 'src'},
        description='iniconfig: brain-dead simple config-ini parsing',
        long_description=readme,
        use_scm_version=True,
        url='http://github.com/RonnyPfannschmidt/iniconfig',
        license='MIT License',
        platforms=['unix', 'linux', 'osx', 'cygwin', 'win32'],
        author='Ronny Pfannschmidt, Holger Krekel',
        author_email=(
            'opensource@ronnypfannschmidt.de, holger.krekel@gmail.com'),
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Operating System :: POSIX',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: MacOS :: MacOS X',
            'Topic :: Software Development :: Libraries',
            'Topic :: Utilities',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 3',
        ],
        include_package_data=True,
        zip_safe=False,
    )

if __name__ == '__main__':
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest
from io import StringIO
from pathlib import Path

import pytest
from mozunit import main
from six import string_types

from mach.base import CommandContext
from mach.registrar import Registrar


@pytest.mark.usefixtures("get_mach", "run_mach")
class TestDispatcher(unittest.TestCase):
    """Tests dispatch related code"""

    def get_parser(self, config=None):
        mach = self.get_mach(Path("basic.py"))

        for provider in Registrar.settings_providers:
            mach.settings.register_provider(provider)

        if config:
            if isinstance(config, string_types):
                config = StringIO(config)
            mach.settings.load_fps([config])

        context = CommandContext(cwd="", settings=mach.settings)
        from mach.main import get_argument_parser

        return get_argument_parser(context)

    def test_command_aliases(self):
        config = """
[alias]
foo = cmd_foo
bar = cmd_bar
baz = cmd_bar --baz
cmd_bar = cmd_bar --baz
"""
        parser = self.get_parser(config=config)

        args = parser.parse_args(["foo"])
        self.assertEqual(args.command, "cmd_foo")

        def assert_bar_baz(argv):
            args = parser.parse_args(argv)
            self.assertEqual(args.command, "cmd_bar")
            self.assertTrue(args.command_args.baz)

        # The following should all result in |cmd_bar --baz|
        assert_bar_baz(["bar", "--baz"])
        assert_bar_baz(["baz"])
        assert_bar_baz(["cmd_bar"])


if __name__ == "__main__":
    main()



# Copyright 2020 Google Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

from six.moves import urllib


def _add_set_cookie(request, value):
    request.extra_headers.append(('Set-Cookie', value))


def web_socket_do_extra_handshake(request):
    components = urllib.parse.urlparse(request.uri)
    command = components[4]

    ONE_DAY_LIFE = 'Max-Age=86400'

    if command == 'set':
        _add_set_cookie(request, '; '.join(['foo=bar', ONE_DAY_LIFE]))
    elif command == 'set_httponly':
        _add_set_cookie(
            request, '; '.join(['httpOnlyFoo=bar', ONE_DAY_LIFE, 'httpOnly']))
    elif command == 'clear':
        _add_set_cookie(request, 'foo=0; Max-Age=0')
        _add_set_cookie(request, 'httpOnlyFoo=0; Max-Age=0')


def web_socket_transfer_data(request):
    pass



#!/usr/bin/env python

import io
import os
import re
import shutil
import tempfile
import threading
import time
import unittest
from unittest import mock

import mozunit
import requests
import six
from mozgeckoprofiler import view_gecko_profile

if six.PY2:
    # Import for Python 2
    from urllib import unquote
else:
    # Import for Python 3
    from urllib.parse import unquote


def access_profiler_link(file_url, response):
    """Attempts to access the profile in a loop for 5 seconds.

    This is run from a separate thread.
    """
    timeout = 5  # seconds
    start = time.time()

    while time.time() - start < timeout:
        # Poll the server to try and get a response.
        result = requests.get(url=file_url)
        if result.ok:
            # Return the text back in a list.
            response[0] = result.text
            return
        time.sleep(0.1)

    response[0] = "Accessing the profiler link timed out after %s seconds" % timeout


class TestViewGeckoProfile(unittest.TestCase):
    """Tests the opening local profiles in the Firefox Profiler."""

    def setUp(self):
        self.firefox_profiler_url = None
        self.thread = None
        self.response = [None]

    def test_view_gecko_profile(self):
        # Create a temporary fake performance profile.
        temp_dir = tempfile.mkdtemp()
        profile_path = os.path.join(temp_dir, "fakeprofile.json")
        with io.open(profile_path, "w") as f:
            f.write("FAKE_PROFILE")

        # Mock the open_new_tab function so that we know when the view_gecko_profile
        # function has done all of its work, and we can assert ressult of the
        # user behavior.
        def mocked_open_new_tab(firefox_profiler_url):
            self.firefox_profiler_url = firefox_profiler_url
            encoded_file_url = firefox_profiler_url.split("/")[-1]
            decoded_file_url = unquote(encoded_file_url)
            # Extract the actual file from the path.
            self.thread = threading.Thread(
                target=access_profiler_link, args=(decoded_file_url, self.response)
            )
            print("firefox_profiler_url %s" % firefox_profiler_url)
            print("encoded_file_url %s" % encoded_file_url)
            print("decoded_file_url %s" % decoded_file_url)
            self.thread.start()

        with mock.patch("webbrowser.open_new_tab", new=mocked_open_new_tab):
            # Run the test
            view_gecko_profile(profile_path)

        self.thread.join()

        # Compare the URLs, but replace the PORT value supplied, as that is dynamic.
        expected_url = (
            "https://profiler.firefox.com/from-url/"
            "http%3A%2F%2F127.0.0.1%3A{PORT}%2Ffakeprofile.json"
        )
        actual_url = re.sub(r"%3A\d+%2F", "%3A{PORT}%2F", self.firefox_profiler_url)

        self.assertEqual(
            actual_url,
            expected_url,
            "The URL generated was correct for the Firefox Profiler.",
        )
        self.assertEqual(
            self.response[0],
            "FAKE_PROFILE",
            "The response from the serve provided the profile contents.",
        )

        shutil.rmtree(temp_dir)


if __name__ == "__main__":
    mozunit.main()



from typing import Any, Union

from . import Attribute, _FilterType

def include(*what: Union[type, Attribute[Any]]) -> _FilterType[Any]: ...
def exclude(*what: Union[type, Attribute[Any]]) -> _FilterType[Any]: ...



# SPDX-License-Identifier: MIT

from __future__ import absolute_import, division, print_function

import pytest

from attr import VersionInfo
from attr._compat import PY2


@pytest.fixture(name="vi")
def fixture_vi():
    return VersionInfo(19, 2, 0, "final")


class TestVersionInfo:
    def test_from_string_no_releaselevel(self, vi):
        """
        If there is no suffix, the releaselevel becomes "final" by default.
        """
        assert vi == VersionInfo._from_version_string("19.2.0")

    def test_suffix_is_preserved(self):
        """
        If there is a suffix, it's preserved.
        """
        assert (
            "dev0"
            == VersionInfo._from_version_string("19.2.0.dev0").releaselevel
        )

    @pytest.mark.skipif(
        PY2, reason="Python 2 is too YOLO to care about comparability."
    )
    @pytest.mark.parametrize("other", [(), (19, 2, 0, "final", "garbage")])
    def test_wrong_len(self, vi, other):
        """
        Comparing with a tuple that has the wrong length raises an error.
        """
        assert vi != other

        with pytest.raises(TypeError):
            vi < other

    @pytest.mark.parametrize("other", [[19, 2, 0, "final"]])
    def test_wrong_type(self, vi, other):
        """
        Only compare to other VersionInfos or tuples.
        """
        assert vi != other

    def test_order(self, vi):
        """
        Ordering works as expected.
        """
        assert vi < (20,)
        assert vi < (19, 2, 1)
        assert vi > (0,)
        assert vi <= (19, 2)
        assert vi >= (19, 2)
        assert vi > (19, 2, 0, "dev0")
        assert vi < (19, 2, 0, "post1")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
from unittest import mock

import pytest
from buildconfig import topsrcdir
from mozunit import main

from mach.site import (
    PIP_NETWORK_INSTALL_RESTRICTED_VIRTUALENVS,
    SitePackagesSource,
    resolve_requirements,
)


@pytest.mark.parametrize(
    "env_native_package_source,env_use_system_python,env_moz_automation,expected",
    [
        ("system", False, False, SitePackagesSource.SYSTEM),
        ("pip", False, False, SitePackagesSource.VENV),
        ("none", False, False, SitePackagesSource.NONE),
        (None, False, False, SitePackagesSource.VENV),
        (None, False, True, SitePackagesSource.NONE),
        (None, True, False, SitePackagesSource.NONE),
        (None, True, True, SitePackagesSource.NONE),
    ],
)
def test_resolve_package_source(
    env_native_package_source, env_use_system_python, env_moz_automation, expected
):
    with mock.patch.dict(
        os.environ,
        {
            "MACH_BUILD_PYTHON_NATIVE_PACKAGE_SOURCE": env_native_package_source or "",
            "MACH_USE_SYSTEM_PYTHON": "1" if env_use_system_python else "",
            "MOZ_AUTOMATION": "1" if env_moz_automation else "",
        },
    ):
        assert SitePackagesSource.for_mach() == expected


def test_all_restricted_sites_dont_have_pypi_requirements():
    for site_name in PIP_NETWORK_INSTALL_RESTRICTED_VIRTUALENVS:
        requirements = resolve_requirements(topsrcdir, site_name)
        assert not requirements.pypi_requirements, (
            'Sites that must be able to operate without "pip install" must not have any '
            f'mandatory "pypi requirements". However, the "{site_name}" site depends on: '
            f"{requirements.pypi_requirements}"
        )


if __name__ == "__main__":
    main()



import asyncio
import contextlib
import logging
import sys
import unittest
import unittest.mock
import warnings

from websockets.exceptions import ConnectionClosed, InvalidState
from websockets.frames import (
    OP_BINARY,
    OP_CLOSE,
    OP_CONT,
    OP_PING,
    OP_PONG,
    OP_TEXT,
    Close,
    CloseCode,
)
from websockets.legacy.framing import Frame
from websockets.legacy.protocol import WebSocketCommonProtocol, broadcast
from websockets.protocol import State

from ..utils import MS
from .utils import AsyncioTestCase


async def async_iterable(iterable):
    for item in iterable:
        yield item


class TransportMock(unittest.mock.Mock):
    """
    Transport mock to control the protocol's inputs and outputs in tests.

    It calls the protocol's connection_made and connection_lost methods like
    actual transports.

    It also calls the protocol's connection_open method to bypass the
    WebSocket handshake.

    To simulate incoming data, tests call the protocol's data_received and
    eof_received methods directly.

    They could also pause_writing and resume_writing to test flow control.

    """

    # This should happen in __init__ but overriding Mock.__init__ is hard.
    def setup_mock(self, loop, protocol):
        self.loop = loop
        self.protocol = protocol
        self._eof = False
        self._closing = False
        # Simulate a successful TCP handshake.
        self.protocol.connection_made(self)
        # Simulate a successful WebSocket handshake.
        self.protocol.connection_open()

    def can_write_eof(self):
        return True

    def write_eof(self):
        # When the protocol half-closes the TCP connection, it expects the
        # other end to close it. Simulate that.
        if not self._eof:
            self.loop.call_soon(self.close)
        self._eof = True

    def close(self):
        # Simulate how actual transports drop the connection.
        if not self._closing:
            self.loop.call_soon(self.protocol.connection_lost, None)
        self._closing = True

    def abort(self):
        # Change this to an `if` if tests call abort() multiple times.
        assert self.protocol.state is not State.CLOSED
        self.loop.call_soon(self.protocol.connection_lost, None)


class CommonTests:
    """
    Mixin that defines most tests but doesn't inherit unittest.TestCase.

    Tests are run by the ServerTests and ClientTests subclasses.

    """

    def setUp(self):
        super().setUp()

        # This logic is encapsulated in a coroutine to prevent it from executing
        # before the event loop is running which causes asyncio.get_event_loop()
        # to raise a DeprecationWarning on Python ≥ 3.10.

        async def create_protocol():
            # Disable pings to make it easier to test what frames are sent exactly.
            return WebSocketCommonProtocol(ping_interval=None)

        self.protocol = self.loop.run_until_complete(create_protocol())
        self.transport = TransportMock()
        self.transport.setup_mock(self.loop, self.protocol)

    def tearDown(self):
        self.transport.close()
        self.loop.run_until_complete(self.protocol.close())
        super().tearDown()

    # Utilities for writing tests.

    def make_drain_slow(self, delay=MS):
        # Process connection_made in order to initialize self.protocol.transport.
        self.run_loop_once()

        original_drain = self.protocol._drain

        async def delayed_drain():
            await asyncio.sleep(delay)
            await original_drain()

        self.protocol._drain = delayed_drain

    close_frame = Frame(
        True,
        OP_CLOSE,
        Close(CloseCode.NORMAL_CLOSURE, "close").serialize(),
    )
    local_close = Frame(
        True,
        OP_CLOSE,
        Close(CloseCode.NORMAL_CLOSURE, "local").serialize(),
    )
    remote_close = Frame(
        True,
        OP_CLOSE,
        Close(CloseCode.NORMAL_CLOSURE, "remote").serialize(),
    )

    def receive_frame(self, frame):
        """
        Make the protocol receive a frame.

        """
        write = self.protocol.data_received
        mask = not self.protocol.is_client
        frame.write(write, mask=mask)

    def receive_eof(self):
        """
        Make the protocol receive the end of the data stream.

        Since ``WebSocketCommonProtocol.eof_received`` returns ``None``, an
        actual transport would close itself after calling it. This function
        emulates that behavior.

        """
        self.protocol.eof_received()
        self.loop.call_soon(self.transport.close)

    def receive_eof_if_client(self):
        """
        Like receive_eof, but only if this is the client side.

        Since the server is supposed to initiate the termination of the TCP
        connection, this method helps making tests work for both sides.

        """
        if self.protocol.is_client:
            self.receive_eof()

    def close_connection(self, code=CloseCode.NORMAL_CLOSURE, reason="close"):
        """
        Execute a closing handshake.

        This puts the connection in the CLOSED state.

        """
        close_frame_data = Close(code, reason).serialize()
        # Prepare the response to the closing handshake from the remote side.
        self.receive_frame(Frame(True, OP_CLOSE, close_frame_data))
        self.receive_eof_if_client()
        # Trigger the closing handshake from the local side and complete it.
        self.loop.run_until_complete(self.protocol.close(code, reason))
        # Empty the outgoing data stream so we can make assertions later on.
        self.assertOneFrameSent(True, OP_CLOSE, close_frame_data)

        assert self.protocol.state is State.CLOSED

    def half_close_connection_local(
        self,
        code=CloseCode.NORMAL_CLOSURE,
        reason="close",
    ):
        """
        Start a closing handshake but do not complete it.

        The main difference with `close_connection` is that the connection is
        left in the CLOSING state until the event loop runs again.

        The current implementation returns a task that must be awaited or
        canceled, else asyncio complains about destroying a pending task.

        """
        close_frame_data = Close(code, reason).serialize()
        # Trigger the closing handshake from the local endpoint.
        close_task = self.loop.create_task(self.protocol.close(code, reason))
        self.run_loop_once()  # write_frame executes
        # Empty the outgoing data stream so we can make assertions later on.
        self.assertOneFrameSent(True, OP_CLOSE, close_frame_data)

        assert self.protocol.state is State.CLOSING

        # Complete the closing sequence at 1ms intervals so the test can run
        # at each point even it goes back to the event loop several times.
        self.loop.call_later(
            MS, self.receive_frame, Frame(True, OP_CLOSE, close_frame_data)
        )
        self.loop.call_later(2 * MS, self.receive_eof_if_client)

        # This task must be awaited or canceled by the caller.
        return close_task

    def half_close_connection_remote(
        self,
        code=CloseCode.NORMAL_CLOSURE,
        reason="close",
    ):
        """
        Receive a closing handshake but do not complete it.

        The main difference with `close_connection` is that the connection is
        left in the CLOSING state until the event loop runs again.

        """
        # On the server side, websockets completes the closing handshake and
        # closes the TCP connection immediately. Yield to the event loop after
        # sending the close frame to run the test while the connection is in
        # the CLOSING state.
        if not self.protocol.is_client:
            self.make_drain_slow()

        close_frame_data = Close(code, reason).serialize()
        # Trigger the closing handshake from the remote endpoint.
        self.receive_frame(Frame(True, OP_CLOSE, close_frame_data))
        self.run_loop_once()  # read_frame executes
        # Empty the outgoing data stream so we can make assertions later on.
        self.assertOneFrameSent(True, OP_CLOSE, close_frame_data)

        assert self.protocol.state is State.CLOSING

        # Complete the closing sequence at 1ms intervals so the test can run
        # at each point even it goes back to the event loop several times.
        self.loop.call_later(2 * MS, self.receive_eof_if_client)

    def process_invalid_frames(self):
        """
        Make the protocol fail quickly after simulating invalid data.

        To achieve this, this function triggers the protocol's eof_received,
        which interrupts pending reads waiting for more data.

        """
        self.run_loop_once()
        self.receive_eof()
        self.loop.run_until_complete(self.protocol.close_connection_task)

    def sent_frames(self):
        """
        Read all frames sent to the transport.

        """
        stream = asyncio.StreamReader(loop=self.loop)

        for (data,), kw in self.transport.write.call_args_list:
            stream.feed_data(data)
        self.transport.write.call_args_list = []
        stream.feed_eof()

        frames = []
        while not stream.at_eof():
            frames.append(
                self.loop.run_until_complete(
                    Frame.read(stream.readexactly, mask=self.protocol.is_client)
                )
            )
        return frames

    def last_sent_frame(self):
        """
        Read the last frame sent to the transport.

        This method assumes that at most one frame was sent. It raises an
        AssertionError otherwise.

        """
        frames = self.sent_frames()
        if frames:
            assert len(frames) == 1
            return frames[0]

    def assertFramesSent(self, *frames):
        self.assertEqual(self.sent_frames(), [Frame(*args) for args in frames])

    def assertOneFrameSent(self, *args):
        self.assertEqual(self.last_sent_frame(), Frame(*args))

    def assertNoFrameSent(self):
        self.assertIsNone(self.last_sent_frame())

    def assertConnectionClosed(self, code, message):
        # The following line guarantees that connection_lost was called.
        self.assertEqual(self.protocol.state, State.CLOSED)
        # A close frame was received.
        self.assertEqual(self.protocol.close_code, code)
        self.assertEqual(self.protocol.close_reason, message)

    def assertConnectionFailed(self, code, message):
        # The following line guarantees that connection_lost was called.
        self.assertEqual(self.protocol.state, State.CLOSED)
        # No close frame was received.
        self.assertEqual(self.protocol.close_code, CloseCode.ABNORMAL_CLOSURE)
        self.assertEqual(self.protocol.close_reason, "")
        # A close frame was sent -- unless the connection was already lost.
        if code == CloseCode.ABNORMAL_CLOSURE:
            self.assertNoFrameSent()
        else:
            self.assertOneFrameSent(True, OP_CLOSE, Close(code, message).serialize())

    @contextlib.contextmanager
    def assertCompletesWithin(self, min_time, max_time):
        t0 = self.loop.time()
        yield
        t1 = self.loop.time()
        dt = t1 - t0
        self.assertGreaterEqual(dt, min_time, f"Too fast: {dt} < {min_time}")
        self.assertLess(dt, max_time, f"Too slow: {dt} >= {max_time}")

    # Test constructor.

    def test_timeout_backwards_compatibility(self):
        async def create_protocol():
            return WebSocketCommonProtocol(ping_interval=None, timeout=5)

        with warnings.catch_warnings(record=True) as recorded:
            warnings.simplefilter("always")
            protocol = self.loop.run_until_complete(create_protocol())

        self.assertEqual(protocol.close_timeout, 5)
        self.assertDeprecationWarnings(recorded, ["rename timeout to close_timeout"])

    def test_loop_backwards_compatibility(self):
        loop = asyncio.new_event_loop()
        self.addCleanup(loop.close)

        with warnings.catch_warnings(record=True) as recorded:
            warnings.simplefilter("always")
            protocol = WebSocketCommonProtocol(ping_interval=None, loop=loop)

        self.assertEqual(protocol.loop, loop)
        self.assertDeprecationWarnings(recorded, ["remove loop argument"])

    # Test public attributes.

    def test_local_address(self):
        get_extra_info = unittest.mock.Mock(return_value=("host", 4312))
        self.transport.get_extra_info = get_extra_info

        self.assertEqual(self.protocol.local_address, ("host", 4312))
        get_extra_info.assert_called_with("sockname")

    def test_local_address_before_connection(self):
        # Emulate the situation before connection_open() runs.
        _transport = self.protocol.transport
        del self.protocol.transport
        try:
            self.assertEqual(self.protocol.local_address, None)
        finally:
            self.protocol.transport = _transport

    def test_remote_address(self):
        get_extra_info = unittest.mock.Mock(return_value=("host", 4312))
        self.transport.get_extra_info = get_extra_info

        self.assertEqual(self.protocol.remote_address, ("host", 4312))
        get_extra_info.assert_called_with("peername")

    def test_remote_address_before_connection(self):
        # Emulate the situation before connection_open() runs.
        _transport = self.protocol.transport
        del self.protocol.transport
        try:
            self.assertEqual(self.protocol.remote_address, None)
        finally:
            self.protocol.transport = _transport

    def test_open(self):
        self.assertTrue(self.protocol.open)
        self.close_connection()
        self.assertFalse(self.protocol.open)

    def test_closed(self):
        self.assertFalse(self.protocol.closed)
        self.close_connection()
        self.assertTrue(self.protocol.closed)

    def test_wait_closed(self):
        wait_closed = self.loop.create_task(self.protocol.wait_closed())
        self.assertFalse(wait_closed.done())
        self.close_connection()
        self.assertTrue(wait_closed.done())

    def test_close_code(self):
        self.close_connection(CloseCode.GOING_AWAY, "Bye!")
        self.assertEqual(self.protocol.close_code, CloseCode.GOING_AWAY)

    def test_close_reason(self):
        self.close_connection(CloseCode.GOING_AWAY, "Bye!")
        self.assertEqual(self.protocol.close_reason, "Bye!")

    def test_close_code_not_set(self):
        self.assertIsNone(self.protocol.close_code)

    def test_close_reason_not_set(self):
        self.assertIsNone(self.protocol.close_reason)

    # Test the recv coroutine.

    def test_recv_text(self):
        self.receive_frame(Frame(True, OP_TEXT, "café".encode("utf-8")))
        data = self.loop.run_until_complete(self.protocol.recv())
        self.assertEqual(data, "café")

    def test_recv_binary(self):
        self.receive_frame(Frame(True, OP_BINARY, b"tea"))
        data = self.loop.run_until_complete(self.protocol.recv())
        self.assertEqual(data, b"tea")

    def test_recv_on_closing_connection_local(self):
        close_task = self.half_close_connection_local()

        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(self.protocol.recv())

        self.loop.run_until_complete(close_task)  # cleanup

    def test_recv_on_closing_connection_remote(self):
        self.half_close_connection_remote()

        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(self.protocol.recv())

    def test_recv_on_closed_connection(self):
        self.close_connection()

        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(self.protocol.recv())

    def test_recv_protocol_error(self):
        self.receive_frame(Frame(True, OP_CONT, "café".encode("utf-8")))
        self.process_invalid_frames()
        self.assertConnectionFailed(CloseCode.PROTOCOL_ERROR, "")

    def test_recv_unicode_error(self):
        self.receive_frame(Frame(True, OP_TEXT, "café".encode("latin-1")))
        self.process_invalid_frames()
        self.assertConnectionFailed(CloseCode.INVALID_DATA, "")

    def test_recv_text_payload_too_big(self):
        self.protocol.max_size = 1024
        self.receive_frame(Frame(True, OP_TEXT, "café".encode("utf-8") * 205))
        self.process_invalid_frames()
        self.assertConnectionFailed(CloseCode.MESSAGE_TOO_BIG, "")

    def test_recv_binary_payload_too_big(self):
        self.protocol.max_size = 1024
        self.receive_frame(Frame(True, OP_BINARY, b"tea" * 342))
        self.process_invalid_frames()
        self.assertConnectionFailed(CloseCode.MESSAGE_TOO_BIG, "")

    def test_recv_text_no_max_size(self):
        self.protocol.max_size = None  # for test coverage
        self.receive_frame(Frame(True, OP_TEXT, "café".encode("utf-8") * 205))
        data = self.loop.run_until_complete(self.protocol.recv())
        self.assertEqual(data, "café" * 205)

    def test_recv_binary_no_max_size(self):
        self.protocol.max_size = None  # for test coverage
        self.receive_frame(Frame(True, OP_BINARY, b"tea" * 342))
        data = self.loop.run_until_complete(self.protocol.recv())
        self.assertEqual(data, b"tea" * 342)

    def test_recv_queue_empty(self):
        recv = self.loop.create_task(self.protocol.recv())
        with self.assertRaises(asyncio.TimeoutError):
            self.loop.run_until_complete(
                asyncio.wait_for(asyncio.shield(recv), timeout=MS)
            )

        self.receive_frame(Frame(True, OP_TEXT, "café".encode("utf-8")))
        data = self.loop.run_until_complete(recv)
        self.assertEqual(data, "café")

    def test_recv_queue_full(self):
        self.protocol.max_queue = 2
        # Test internals because it's hard to verify buffers from the outside.
        self.assertEqual(list(self.protocol.messages), [])

        self.receive_frame(Frame(True, OP_TEXT, "café".encode("utf-8")))
        self.run_loop_once()
        self.assertEqual(list(self.protocol.messages), ["café"])

        self.receive_frame(Frame(True, OP_BINARY, b"tea"))
        self.run_loop_once()
        self.assertEqual(list(self.protocol.messages), ["café", b"tea"])

        self.receive_frame(Frame(True, OP_BINARY, b"milk"))
        self.run_loop_once()
        self.assertEqual(list(self.protocol.messages), ["café", b"tea"])

        self.loop.run_until_complete(self.protocol.recv())
        self.run_loop_once()
        self.assertEqual(list(self.protocol.messages), [b"tea", b"milk"])

        self.loop.run_until_complete(self.protocol.recv())
        self.run_loop_once()
        self.assertEqual(list(self.protocol.messages), [b"milk"])

        self.loop.run_until_complete(self.protocol.recv())
        self.run_loop_once()
        self.assertEqual(list(self.protocol.messages), [])

    def test_recv_queue_no_limit(self):
        self.protocol.max_queue = None

        for _ in range(100):
            self.receive_frame(Frame(True, OP_TEXT, "café".encode("utf-8")))
            self.run_loop_once()

        # Incoming message queue can contain at least 100 messages.
        self.assertEqual(list(self.protocol.messages), ["café"] * 100)

        for _ in range(100):
            self.loop.run_until_complete(self.protocol.recv())

        self.assertEqual(list(self.protocol.messages), [])

    def test_recv_other_error(self):
        async def read_message():
            raise Exception("BOOM")

        self.protocol.read_message = read_message
        self.process_invalid_frames()
        self.assertConnectionFailed(CloseCode.INTERNAL_ERROR, "")

    def test_recv_canceled(self):
        recv = self.loop.create_task(self.protocol.recv())
        self.loop.call_soon(recv.cancel)

        with self.assertRaises(asyncio.CancelledError):
            self.loop.run_until_complete(recv)

        # The next frame doesn't disappear in a vacuum (it used to).
        self.receive_frame(Frame(True, OP_TEXT, "café".encode("utf-8")))
        data = self.loop.run_until_complete(self.protocol.recv())
        self.assertEqual(data, "café")

    def test_recv_canceled_race_condition(self):
        recv = self.loop.create_task(
            asyncio.wait_for(self.protocol.recv(), timeout=0.000_001)
        )
        self.loop.call_soon(
            self.receive_frame, Frame(True, OP_TEXT, "café".encode("utf-8"))
        )

        with self.assertRaises(asyncio.TimeoutError):
            self.loop.run_until_complete(recv)

        # The previous frame doesn't disappear in a vacuum (it used to).
        self.receive_frame(Frame(True, OP_TEXT, "tea".encode("utf-8")))
        data = self.loop.run_until_complete(self.protocol.recv())
        # If we're getting "tea" there, it means "café" was swallowed (ha, ha).
        self.assertEqual(data, "café")

    def test_recv_when_transfer_data_cancelled(self):
        # Clog incoming queue.
        self.protocol.max_queue = 1
        self.receive_frame(Frame(True, OP_TEXT, "café".encode("utf-8")))
        self.receive_frame(Frame(True, OP_BINARY, b"tea"))
        self.run_loop_once()

        # Flow control kicks in (check with an implementation detail).
        self.assertFalse(self.protocol._put_message_waiter.done())

        # Schedule recv().
        recv = self.loop.create_task(self.protocol.recv())

        # Cancel transfer_data_task (again, implementation detail).
        self.protocol.fail_connection()
        self.run_loop_once()
        self.assertTrue(self.protocol.transfer_data_task.cancelled())

        # recv() completes properly.
        self.assertEqual(self.loop.run_until_complete(recv), "café")

    def test_recv_prevents_concurrent_calls(self):
        recv = self.loop.create_task(self.protocol.recv())

        with self.assertRaises(RuntimeError) as raised:
            self.loop.run_until_complete(self.protocol.recv())
        self.assertEqual(
            str(raised.exception),
            "cannot call recv while another coroutine "
            "is already waiting for the next message",
        )
        recv.cancel()

    # Test the send coroutine.

    def test_send_text(self):
        self.loop.run_until_complete(self.protocol.send("café"))
        self.assertOneFrameSent(True, OP_TEXT, "café".encode("utf-8"))

    def test_send_binary(self):
        self.loop.run_until_complete(self.protocol.send(b"tea"))
        self.assertOneFrameSent(True, OP_BINARY, b"tea")

    def test_send_binary_from_bytearray(self):
        self.loop.run_until_complete(self.protocol.send(bytearray(b"tea")))
        self.assertOneFrameSent(True, OP_BINARY, b"tea")

    def test_send_binary_from_memoryview(self):
        self.loop.run_until_complete(self.protocol.send(memoryview(b"tea")))
        self.assertOneFrameSent(True, OP_BINARY, b"tea")

    def test_send_dict(self):
        with self.assertRaises(TypeError):
            self.loop.run_until_complete(self.protocol.send({"not": "encoded"}))
        self.assertNoFrameSent()

    def test_send_type_error(self):
        with self.assertRaises(TypeError):
            self.loop.run_until_complete(self.protocol.send(42))
        self.assertNoFrameSent()

    def test_send_iterable_text(self):
        self.loop.run_until_complete(self.protocol.send(["ca", "fé"]))
        self.assertFramesSent(
            (False, OP_TEXT, "ca".encode("utf-8")),
            (False, OP_CONT, "fé".encode("utf-8")),
            (True, OP_CONT, "".encode("utf-8")),
        )

    def test_send_iterable_binary(self):
        self.loop.run_until_complete(self.protocol.send([b"te", b"a"]))
        self.assertFramesSent(
            (False, OP_BINARY, b"te"), (False, OP_CONT, b"a"), (True, OP_CONT, b"")
        )

    def test_send_iterable_binary_from_bytearray(self):
        self.loop.run_until_complete(
            self.protocol.send([bytearray(b"te"), bytearray(b"a")])
        )
        self.assertFramesSent(
            (False, OP_BINARY, b"te"), (False, OP_CONT, b"a"), (True, OP_CONT, b"")
        )

    def test_send_iterable_binary_from_memoryview(self):
        self.loop.run_until_complete(
            self.protocol.send([memoryview(b"te"), memoryview(b"a")])
        )
        self.assertFramesSent(
            (False, OP_BINARY, b"te"), (False, OP_CONT, b"a"), (True, OP_CONT, b"")
        )

    def test_send_empty_iterable(self):
        self.loop.run_until_complete(self.protocol.send([]))
        self.assertNoFrameSent()

    def test_send_iterable_type_error(self):
        with self.assertRaises(TypeError):
            self.loop.run_until_complete(self.protocol.send([42]))
        self.assertNoFrameSent()

    def test_send_iterable_mixed_type_error(self):
        with self.assertRaises(TypeError):
            self.loop.run_until_complete(self.protocol.send(["café", b"tea"]))
        self.assertFramesSent(
            (False, OP_TEXT, "café".encode("utf-8")),
            (True, OP_CLOSE, Close(CloseCode.INTERNAL_ERROR, "").serialize()),
        )

    def test_send_iterable_prevents_concurrent_send(self):
        self.make_drain_slow(2 * MS)

        async def send_iterable():
            await self.protocol.send(["ca", "fé"])

        async def send_concurrent():
            await asyncio.sleep(MS)
            await self.protocol.send(b"tea")

        async def run_concurrently():
            await asyncio.gather(
                send_iterable(),
                send_concurrent(),
            )

        self.loop.run_until_complete(run_concurrently())

        self.assertFramesSent(
            (False, OP_TEXT, "ca".encode("utf-8")),
            (False, OP_CONT, "fé".encode("utf-8")),
            (True, OP_CONT, "".encode("utf-8")),
            (True, OP_BINARY, b"tea"),
        )

    def test_send_async_iterable_text(self):
        self.loop.run_until_complete(self.protocol.send(async_iterable(["ca", "fé"])))
        self.assertFramesSent(
            (False, OP_TEXT, "ca".encode("utf-8")),
            (False, OP_CONT, "fé".encode("utf-8")),
            (True, OP_CONT, "".encode("utf-8")),
        )

    def test_send_async_iterable_binary(self):
        self.loop.run_until_complete(self.protocol.send(async_iterable([b"te", b"a"])))
        self.assertFramesSent(
            (False, OP_BINARY, b"te"), (False, OP_CONT, b"a"), (True, OP_CONT, b"")
        )

    def test_send_async_iterable_binary_from_bytearray(self):
        self.loop.run_until_complete(
            self.protocol.send(async_iterable([bytearray(b"te"), bytearray(b"a")]))
        )
        self.assertFramesSent(
            (False, OP_BINARY, b"te"), (False, OP_CONT, b"a"), (True, OP_CONT, b"")
        )

    def test_send_async_iterable_binary_from_memoryview(self):
        self.loop.run_until_complete(
            self.protocol.send(async_iterable([memoryview(b"te"), memoryview(b"a")]))
        )
        self.assertFramesSent(
            (False, OP_BINARY, b"te"), (False, OP_CONT, b"a"), (True, OP_CONT, b"")
        )

    def test_send_empty_async_iterable(self):
        self.loop.run_until_complete(self.protocol.send(async_iterable([])))
        self.assertNoFrameSent()

    def test_send_async_iterable_type_error(self):
        with self.assertRaises(TypeError):
            self.loop.run_until_complete(self.protocol.send(async_iterable([42])))
        self.assertNoFrameSent()

    def test_send_async_iterable_mixed_type_error(self):
        with self.assertRaises(TypeError):
            self.loop.run_until_complete(
                self.protocol.send(async_iterable(["café", b"tea"]))
            )
        self.assertFramesSent(
            (False, OP_TEXT, "café".encode("utf-8")),
            (True, OP_CLOSE, Close(CloseCode.INTERNAL_ERROR, "").serialize()),
        )

    def test_send_async_iterable_prevents_concurrent_send(self):
        self.make_drain_slow(2 * MS)

        async def send_async_iterable():
            await self.protocol.send(async_iterable(["ca", "fé"]))

        async def send_concurrent():
            await asyncio.sleep(MS)
            await self.protocol.send(b"tea")

        async def run_concurrently():
            await asyncio.gather(
                send_async_iterable(),
                send_concurrent(),
            )

        self.loop.run_until_complete(run_concurrently())

        self.assertFramesSent(
            (False, OP_TEXT, "ca".encode("utf-8")),
            (False, OP_CONT, "fé".encode("utf-8")),
            (True, OP_CONT, "".encode("utf-8")),
            (True, OP_BINARY, b"tea"),
        )

    def test_send_on_closing_connection_local(self):
        close_task = self.half_close_connection_local()

        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(self.protocol.send("foobar"))

        self.assertNoFrameSent()

        self.loop.run_until_complete(close_task)  # cleanup

    def test_send_on_closing_connection_remote(self):
        self.half_close_connection_remote()

        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(self.protocol.send("foobar"))

        self.assertNoFrameSent()

    def test_send_on_closed_connection(self):
        self.close_connection()

        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(self.protocol.send("foobar"))

        self.assertNoFrameSent()

    # Test the ping coroutine.

    def test_ping_default(self):
        self.loop.run_until_complete(self.protocol.ping())
        # With our testing tools, it's more convenient to extract the expected
        # ping data from the library's internals than from the frame sent.
        ping_data = next(iter(self.protocol.pings))
        self.assertIsInstance(ping_data, bytes)
        self.assertEqual(len(ping_data), 4)
        self.assertOneFrameSent(True, OP_PING, ping_data)

    def test_ping_text(self):
        self.loop.run_until_complete(self.protocol.ping("café"))
        self.assertOneFrameSent(True, OP_PING, "café".encode("utf-8"))

    def test_ping_binary(self):
        self.loop.run_until_complete(self.protocol.ping(b"tea"))
        self.assertOneFrameSent(True, OP_PING, b"tea")

    def test_ping_binary_from_bytearray(self):
        self.loop.run_until_complete(self.protocol.ping(bytearray(b"tea")))
        self.assertOneFrameSent(True, OP_PING, b"tea")

    def test_ping_binary_from_memoryview(self):
        self.loop.run_until_complete(self.protocol.ping(memoryview(b"tea")))
        self.assertOneFrameSent(True, OP_PING, b"tea")

    def test_ping_type_error(self):
        with self.assertRaises(TypeError):
            self.loop.run_until_complete(self.protocol.ping(42))
        self.assertNoFrameSent()

    def test_ping_on_closing_connection_local(self):
        close_task = self.half_close_connection_local()

        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(self.protocol.ping())

        self.assertNoFrameSent()

        self.loop.run_until_complete(close_task)  # cleanup

    def test_ping_on_closing_connection_remote(self):
        self.half_close_connection_remote()

        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(self.protocol.ping())

        self.assertNoFrameSent()

    def test_ping_on_closed_connection(self):
        self.close_connection()

        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(self.protocol.ping())

        self.assertNoFrameSent()

    # Test the pong coroutine.

    def test_pong_default(self):
        self.loop.run_until_complete(self.protocol.pong())
        self.assertOneFrameSent(True, OP_PONG, b"")

    def test_pong_text(self):
        self.loop.run_until_complete(self.protocol.pong("café"))
        self.assertOneFrameSent(True, OP_PONG, "café".encode("utf-8"))

    def test_pong_binary(self):
        self.loop.run_until_complete(self.protocol.pong(b"tea"))
        self.assertOneFrameSent(True, OP_PONG, b"tea")

    def test_pong_binary_from_bytearray(self):
        self.loop.run_until_complete(self.protocol.pong(bytearray(b"tea")))
        self.assertOneFrameSent(True, OP_PONG, b"tea")

    def test_pong_binary_from_memoryview(self):
        self.loop.run_until_complete(self.protocol.pong(memoryview(b"tea")))
        self.assertOneFrameSent(True, OP_PONG, b"tea")

    def test_pong_type_error(self):
        with self.assertRaises(TypeError):
            self.loop.run_until_complete(self.protocol.pong(42))
        self.assertNoFrameSent()

    def test_pong_on_closing_connection_local(self):
        close_task = self.half_close_connection_local()

        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(self.protocol.pong())

        self.assertNoFrameSent()

        self.loop.run_until_complete(close_task)  # cleanup

    def test_pong_on_closing_connection_remote(self):
        self.half_close_connection_remote()

        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(self.protocol.pong())

        self.assertNoFrameSent()

    def test_pong_on_closed_connection(self):
        self.close_connection()

        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(self.protocol.pong())

        self.assertNoFrameSent()

    # Test the protocol's logic for acknowledging pings with pongs.

    def test_answer_ping(self):
        self.receive_frame(Frame(True, OP_PING, b"test"))
        self.run_loop_once()
        self.assertOneFrameSent(True, OP_PONG, b"test")

    def test_answer_ping_does_not_crash_if_connection_closing(self):
        close_task = self.half_close_connection_local()

        self.receive_frame(Frame(True, OP_PING, b"test"))
        self.run_loop_once()

        with self.assertNoLogs():
            self.loop.run_until_complete(self.protocol.close())

        self.loop.run_until_complete(close_task)  # cleanup

    def test_answer_ping_does_not_crash_if_connection_closed(self):
        self.make_drain_slow()
        # Drop the connection right after receiving a ping frame,
        # which prevents responding with a pong frame properly.
        self.receive_frame(Frame(True, OP_PING, b"test"))
        self.receive_eof()
        self.run_loop_once()

        with self.assertNoLogs():
            self.loop.run_until_complete(self.protocol.close())

    def test_ignore_pong(self):
        self.receive_frame(Frame(True, OP_PONG, b"test"))
        self.run_loop_once()
        self.assertNoFrameSent()

    def test_acknowledge_ping(self):
        pong_waiter = self.loop.run_until_complete(self.protocol.ping())
        self.assertFalse(pong_waiter.done())
        ping_frame = self.last_sent_frame()
        pong_frame = Frame(True, OP_PONG, ping_frame.data)
        self.receive_frame(pong_frame)
        self.run_loop_once()
        self.run_loop_once()
        self.assertTrue(pong_waiter.done())

    def test_abort_ping(self):
        pong_waiter = self.loop.run_until_complete(self.protocol.ping())
        # Remove the frame from the buffer, else close_connection() complains.
        self.last_sent_frame()
        self.assertFalse(pong_waiter.done())
        self.close_connection()
        self.assertTrue(pong_waiter.done())
        self.assertIsInstance(pong_waiter.exception(), ConnectionClosed)

    def test_abort_ping_does_not_log_exception_if_not_retreived(self):
        self.loop.run_until_complete(self.protocol.ping())
        # Get the internal Future, which isn't directly returned by ping().
        ((pong_waiter, _timestamp),) = self.protocol.pings.values()
        # Remove the frame from the buffer, else close_connection() complains.
        self.last_sent_frame()
        self.close_connection()
        # Check a private attribute, for lack of a better solution.
        self.assertFalse(pong_waiter._log_traceback)

    def test_acknowledge_previous_pings(self):
        pings = [
            (self.loop.run_until_complete(self.protocol.ping()), self.last_sent_frame())
            for i in range(3)
        ]
        # Unsolicited pong doesn't acknowledge pings
        self.receive_frame(Frame(True, OP_PONG, b""))
        self.run_loop_once()
        self.run_loop_once()
        self.assertFalse(pings[0][0].done())
        self.assertFalse(pings[1][0].done())
        self.assertFalse(pings[2][0].done())
        # Pong acknowledges all previous pings
        self.receive_frame(Frame(True, OP_PONG, pings[1][1].data))
        self.run_loop_once()
        self.run_loop_once()
        self.assertTrue(pings[0][0].done())
        self.assertTrue(pings[1][0].done())
        self.assertFalse(pings[2][0].done())

    def test_acknowledge_aborted_ping(self):
        pong_waiter = self.loop.run_until_complete(self.protocol.ping())
        ping_frame = self.last_sent_frame()
        # Clog incoming queue. This lets connection_lost() abort pending pings
        # with a ConnectionClosed exception before transfer_data_task
        # terminates and close_connection cancels keepalive_ping_task.
        self.protocol.max_queue = 1
        self.receive_frame(Frame(True, OP_TEXT, b"1"))
        self.receive_frame(Frame(True, OP_TEXT, b"2"))
        # Add pong frame to the queue.
        pong_frame = Frame(True, OP_PONG, ping_frame.data)
        self.receive_frame(pong_frame)
        # Connection drops.
        self.receive_eof()
        self.loop.run_until_complete(self.protocol.wait_closed())
        # Ping receives a ConnectionClosed exception.
        with self.assertRaises(ConnectionClosed):
            pong_waiter.result()

        # transfer_data doesn't crash, which would be logged.
        with self.assertNoLogs():
            # Unclog incoming queue.
            self.loop.run_until_complete(self.protocol.recv())
            self.loop.run_until_complete(self.protocol.recv())

    def test_canceled_ping(self):
        pong_waiter = self.loop.run_until_complete(self.protocol.ping())
        ping_frame = self.last_sent_frame()
        pong_waiter.cancel()
        pong_frame = Frame(True, OP_PONG, ping_frame.data)
        self.receive_frame(pong_frame)
        self.run_loop_once()
        self.run_loop_once()
        self.assertTrue(pong_waiter.cancelled())

    def test_duplicate_ping(self):
        self.loop.run_until_complete(self.protocol.ping(b"foobar"))
        self.assertOneFrameSent(True, OP_PING, b"foobar")
        with self.assertRaises(RuntimeError):
            self.loop.run_until_complete(self.protocol.ping(b"foobar"))
        self.assertNoFrameSent()

    # Test the protocol's logic for measuring latency

    def test_record_latency_on_pong(self):
        self.assertEqual(self.protocol.latency, 0)
        self.loop.run_until_complete(self.protocol.ping(b"test"))
        self.receive_frame(Frame(True, OP_PONG, b"test"))
        self.run_loop_once()
        self.assertGreater(self.protocol.latency, 0)

    def test_return_latency_on_pong(self):
        pong_waiter = self.loop.run_until_complete(self.protocol.ping())
        ping_frame = self.last_sent_frame()
        pong_frame = Frame(True, OP_PONG, ping_frame.data)
        self.receive_frame(pong_frame)
        latency = self.loop.run_until_complete(pong_waiter)
        self.assertGreater(latency, 0)

    # Test the protocol's logic for rebuilding fragmented messages.

    def test_fragmented_text(self):
        self.receive_frame(Frame(False, OP_TEXT, "ca".encode("utf-8")))
        self.receive_frame(Frame(True, OP_CONT, "fé".encode("utf-8")))
        data = self.loop.run_until_complete(self.protocol.recv())
        self.assertEqual(data, "café")

    def test_fragmented_binary(self):
        self.receive_frame(Frame(False, OP_BINARY, b"t"))
        self.receive_frame(Frame(False, OP_CONT, b"e"))
        self.receive_frame(Frame(True, OP_CONT, b"a"))
        data = self.loop.run_until_complete(self.protocol.recv())
        self.assertEqual(data, b"tea")

    def test_fragmented_text_payload_too_big(self):
        self.protocol.max_size = 1024
        self.receive_frame(Frame(False, OP_TEXT, "café".encode("utf-8") * 100))
        self.receive_frame(Frame(True, OP_CONT, "café".encode("utf-8") * 105))
        self.process_invalid_frames()
        self.assertConnectionFailed(CloseCode.MESSAGE_TOO_BIG, "")

    def test_fragmented_binary_payload_too_big(self):
        self.protocol.max_size = 1024
        self.receive_frame(Frame(False, OP_BINARY, b"tea" * 171))
        self.receive_frame(Frame(True, OP_CONT, b"tea" * 171))
        self.process_invalid_frames()
        self.assertConnectionFailed(CloseCode.MESSAGE_TOO_BIG, "")

    def test_fragmented_text_no_max_size(self):
        self.protocol.max_size = None  # for test coverage
        self.receive_frame(Frame(False, OP_TEXT, "café".encode("utf-8") * 100))
        self.receive_frame(Frame(True, OP_CONT, "café".encode("utf-8") * 105))
        data = self.loop.run_until_complete(self.protocol.recv())
        self.assertEqual(data, "café" * 205)

    def test_fragmented_binary_no_max_size(self):
        self.protocol.max_size = None  # for test coverage
        self.receive_frame(Frame(False, OP_BINARY, b"tea" * 171))
        self.receive_frame(Frame(True, OP_CONT, b"tea" * 171))
        data = self.loop.run_until_complete(self.protocol.recv())
        self.assertEqual(data, b"tea" * 342)

    def test_control_frame_within_fragmented_text(self):
        self.receive_frame(Frame(False, OP_TEXT, "ca".encode("utf-8")))
        self.receive_frame(Frame(True, OP_PING, b""))
        self.receive_frame(Frame(True, OP_CONT, "fé".encode("utf-8")))
        data = self.loop.run_until_complete(self.protocol.recv())
        self.assertEqual(data, "café")
        self.assertOneFrameSent(True, OP_PONG, b"")

    def test_unterminated_fragmented_text(self):
        self.receive_frame(Frame(False, OP_TEXT, "ca".encode("utf-8")))
        # Missing the second part of the fragmented frame.
        self.receive_frame(Frame(True, OP_BINARY, b"tea"))
        self.process_invalid_frames()
        self.assertConnectionFailed(CloseCode.PROTOCOL_ERROR, "")

    def test_close_handshake_in_fragmented_text(self):
        self.receive_frame(Frame(False, OP_TEXT, "ca".encode("utf-8")))
        self.receive_frame(Frame(True, OP_CLOSE, b""))
        self.process_invalid_frames()
        # The RFC may have overlooked this case: it says that control frames
        # can be interjected in the middle of a fragmented message and that a
        # close frame must be echoed. Even though there's an unterminated
        # message, technically, the closing handshake was successful.
        self.assertConnectionClosed(CloseCode.NO_STATUS_RCVD, "")

    def test_connection_close_in_fragmented_text(self):
        self.receive_frame(Frame(False, OP_TEXT, "ca".encode("utf-8")))
        self.process_invalid_frames()
        self.assertConnectionFailed(CloseCode.ABNORMAL_CLOSURE, "")

    # Test miscellaneous code paths to ensure full coverage.

    def test_connection_lost(self):
        # Test calling connection_lost without going through close_connection.
        self.protocol.connection_lost(None)

        self.assertConnectionFailed(CloseCode.ABNORMAL_CLOSURE, "")

    def test_ensure_open_before_opening_handshake(self):
        # Simulate a bug by forcibly reverting the protocol state.
        self.protocol.state = State.CONNECTING

        with self.assertRaises(InvalidState):
            self.loop.run_until_complete(self.protocol.ensure_open())

    def test_ensure_open_during_unclean_close(self):
        # Process connection_made in order to start transfer_data_task.
        self.run_loop_once()

        # Ensure the test terminates quickly.
        self.loop.call_later(MS, self.receive_eof_if_client)

        # Simulate the case when close() times out sending a close frame.
        self.protocol.fail_connection()

        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(self.protocol.ensure_open())

    def test_legacy_recv(self):
        # By default legacy_recv in disabled.
        self.assertEqual(self.protocol.legacy_recv, False)

        self.close_connection()

        # Enable legacy_recv.
        self.protocol.legacy_recv = True

        # Now recv() returns None instead of raising ConnectionClosed.
        self.assertIsNone(self.loop.run_until_complete(self.protocol.recv()))

    def test_connection_closed_attributes(self):
        self.close_connection()

        with self.assertRaises(ConnectionClosed) as context:
            self.loop.run_until_complete(self.protocol.recv())

        connection_closed_exc = context.exception
        self.assertEqual(connection_closed_exc.code, CloseCode.NORMAL_CLOSURE)
        self.assertEqual(connection_closed_exc.reason, "close")

    # Test the protocol logic for sending keepalive pings.

    def restart_protocol_with_keepalive_ping(
        self,
        ping_interval=3 * MS,
        ping_timeout=3 * MS,
    ):
        initial_protocol = self.protocol

        # copied from tearDown

        self.transport.close()
        self.loop.run_until_complete(self.protocol.close())

        # copied from setUp, but enables keepalive pings

        async def create_protocol():
            return WebSocketCommonProtocol(
                ping_interval=ping_interval,
                ping_timeout=ping_timeout,
            )

        self.protocol = self.loop.run_until_complete(create_protocol())

        self.transport = TransportMock()
        self.transport.setup_mock(self.loop, self.protocol)
        self.protocol.is_client = initial_protocol.is_client
        self.protocol.side = initial_protocol.side

    def test_keepalive_ping(self):
        self.restart_protocol_with_keepalive_ping()

        # Ping is sent at 3ms and acknowledged at 4ms.
        self.loop.run_until_complete(asyncio.sleep(4 * MS))
        (ping_1,) = tuple(self.protocol.pings)
        self.assertOneFrameSent(True, OP_PING, ping_1)
        self.receive_frame(Frame(True, OP_PONG, ping_1))

        # Next ping is sent at 7ms.
        self.loop.run_until_complete(asyncio.sleep(4 * MS))
        (ping_2,) = tuple(self.protocol.pings)
        self.assertOneFrameSent(True, OP_PING, ping_2)

        # The keepalive ping task goes on.
        self.assertFalse(self.protocol.keepalive_ping_task.done())

    def test_keepalive_ping_not_acknowledged_closes_connection(self):
        self.restart_protocol_with_keepalive_ping()

        # Ping is sent at 3ms and not acknowledged.
        self.loop.run_until_complete(asyncio.sleep(4 * MS))
        (ping_1,) = tuple(self.protocol.pings)
        self.assertOneFrameSent(True, OP_PING, ping_1)

        # Connection is closed at 6ms.
        self.loop.run_until_complete(asyncio.sleep(4 * MS))
        self.assertOneFrameSent(
            True,
            OP_CLOSE,
            Close(CloseCode.INTERNAL_ERROR, "keepalive ping timeout").serialize(),
        )

        # The keepalive ping task is complete.
        self.assertEqual(self.protocol.keepalive_ping_task.result(), None)

    def test_keepalive_ping_stops_when_connection_closing(self):
        self.restart_protocol_with_keepalive_ping()
        close_task = self.half_close_connection_local()

        # No ping sent at 3ms because the closing handshake is in progress.
        self.loop.run_until_complete(asyncio.sleep(4 * MS))
        self.assertNoFrameSent()

        # The keepalive ping task terminated.
        self.assertTrue(self.protocol.keepalive_ping_task.cancelled())

        self.loop.run_until_complete(close_task)  # cleanup

    def test_keepalive_ping_stops_when_connection_closed(self):
        self.restart_protocol_with_keepalive_ping()
        self.close_connection()

        # The keepalive ping task terminated.
        self.assertTrue(self.protocol.keepalive_ping_task.cancelled())

    def test_keepalive_ping_does_not_crash_when_connection_lost(self):
        self.restart_protocol_with_keepalive_ping()
        # Clog incoming queue. This lets connection_lost() abort pending pings
        # with a ConnectionClosed exception before transfer_data_task
        # terminates and close_connection cancels keepalive_ping_task.
        self.protocol.max_queue = 1
        self.receive_frame(Frame(True, OP_TEXT, b"1"))
        self.receive_frame(Frame(True, OP_TEXT, b"2"))
        # Ping is sent at 3ms.
        self.loop.run_until_complete(asyncio.sleep(4 * MS))
        ((pong_waiter, _timestamp),) = self.protocol.pings.values()
        # Connection drops.
        self.receive_eof()
        self.loop.run_until_complete(self.protocol.wait_closed())

        # The ping waiter receives a ConnectionClosed exception.
        with self.assertRaises(ConnectionClosed):
            pong_waiter.result()
        # The keepalive ping task terminated properly.
        self.assertIsNone(self.protocol.keepalive_ping_task.result())

        # Unclog incoming queue to terminate the test quickly.
        self.loop.run_until_complete(self.protocol.recv())
        self.loop.run_until_complete(self.protocol.recv())

    def test_keepalive_ping_with_no_ping_interval(self):
        self.restart_protocol_with_keepalive_ping(ping_interval=None)

        # No ping is sent at 3ms.
        self.loop.run_until_complete(asyncio.sleep(4 * MS))
        self.assertNoFrameSent()

    def test_keepalive_ping_with_no_ping_timeout(self):
        self.restart_protocol_with_keepalive_ping(ping_timeout=None)

        # Ping is sent at 3ms and not acknowledged.
        self.loop.run_until_complete(asyncio.sleep(4 * MS))
        (ping_1,) = tuple(self.protocol.pings)
        self.assertOneFrameSent(True, OP_PING, ping_1)

        # Next ping is sent at 7ms anyway.
        self.loop.run_until_complete(asyncio.sleep(4 * MS))
        ping_1_again, ping_2 = tuple(self.protocol.pings)
        self.assertEqual(ping_1, ping_1_again)
        self.assertOneFrameSent(True, OP_PING, ping_2)

        # The keepalive ping task goes on.
        self.assertFalse(self.protocol.keepalive_ping_task.done())

    def test_keepalive_ping_unexpected_error(self):
        self.restart_protocol_with_keepalive_ping()

        async def ping():
            raise Exception("BOOM")

        self.protocol.ping = ping

        # The keepalive ping task fails when sending a ping at 3ms.
        self.loop.run_until_complete(asyncio.sleep(4 * MS))

        # The keepalive ping task is complete.
        # It logs and swallows the exception.
        self.assertEqual(self.protocol.keepalive_ping_task.result(), None)

    # Test the protocol logic for closing the connection.

    def test_local_close(self):
        # Emulate how the remote endpoint answers the closing handshake.
        self.loop.call_later(MS, self.receive_frame, self.close_frame)
        self.loop.call_later(MS, self.receive_eof_if_client)

        # Run the closing handshake.
        self.loop.run_until_complete(self.protocol.close(reason="close"))

        self.assertConnectionClosed(CloseCode.NORMAL_CLOSURE, "close")
        self.assertOneFrameSent(*self.close_frame)

        # Closing the connection again is a no-op.
        self.loop.run_until_complete(self.protocol.close(reason="oh noes!"))

        self.assertConnectionClosed(CloseCode.NORMAL_CLOSURE, "close")
        self.assertNoFrameSent()

    def test_remote_close(self):
        # Emulate how the remote endpoint initiates the closing handshake.
        self.loop.call_later(MS, self.receive_frame, self.close_frame)
        self.loop.call_later(MS, self.receive_eof_if_client)

        # Wait for some data in order to process the handshake.
        # After recv() raises ConnectionClosed, the connection is closed.
        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(self.protocol.recv())

        self.assertConnectionClosed(CloseCode.NORMAL_CLOSURE, "close")
        self.assertOneFrameSent(*self.close_frame)

        # Closing the connection again is a no-op.
        self.loop.run_until_complete(self.protocol.close(reason="oh noes!"))

        self.assertConnectionClosed(CloseCode.NORMAL_CLOSURE, "close")
        self.assertNoFrameSent()

    def test_remote_close_and_connection_lost(self):
        self.make_drain_slow()
        # Drop the connection right after receiving a close frame,
        # which prevents echoing the close frame properly.
        self.receive_frame(self.close_frame)
        self.receive_eof()
        self.run_loop_once()

        with self.assertNoLogs():
            self.loop.run_until_complete(self.protocol.close(reason="oh noes!"))

        self.assertConnectionClosed(CloseCode.NORMAL_CLOSURE, "close")
        self.assertOneFrameSent(*self.close_frame)

    def test_simultaneous_close(self):
        # Receive the incoming close frame right after self.protocol.close()
        # starts executing. This reproduces the error described in:
        # https://github.com/python-websockets/websockets/issues/339
        self.loop.call_soon(self.receive_frame, self.remote_close)
        self.loop.call_soon(self.receive_eof_if_client)
        self.run_loop_once()

        self.loop.run_until_complete(self.protocol.close(reason="local"))

        self.assertConnectionClosed(CloseCode.NORMAL_CLOSURE, "remote")
        # The current implementation sends a close frame in response to the
        # close frame received from the remote end. It skips the close frame
        # that should be sent as a result of calling close().
        self.assertOneFrameSent(*self.remote_close)

    def test_close_preserves_incoming_frames(self):
        self.receive_frame(Frame(True, OP_TEXT, b"hello"))
        self.run_loop_once()

        self.loop.call_later(MS, self.receive_frame, self.close_frame)
        self.loop.call_later(MS, self.receive_eof_if_client)
        self.loop.run_until_complete(self.protocol.close(reason="close"))

        self.assertConnectionClosed(CloseCode.NORMAL_CLOSURE, "close")
        self.assertOneFrameSent(*self.close_frame)

        next_message = self.loop.run_until_complete(self.protocol.recv())
        self.assertEqual(next_message, "hello")

    def test_close_protocol_error(self):
        invalid_close_frame = Frame(True, OP_CLOSE, b"\x00")
        self.receive_frame(invalid_close_frame)
        self.receive_eof_if_client()
        self.run_loop_once()
        self.loop.run_until_complete(self.protocol.close(reason="close"))

        self.assertConnectionFailed(CloseCode.PROTOCOL_ERROR, "")

    def test_close_connection_lost(self):
        self.receive_eof()
        self.run_loop_once()
        self.loop.run_until_complete(self.protocol.close(reason="close"))

        self.assertConnectionFailed(CloseCode.ABNORMAL_CLOSURE, "")

    def test_local_close_during_recv(self):
        recv = self.loop.create_task(self.protocol.recv())

        self.loop.call_later(MS, self.receive_frame, self.close_frame)
        self.loop.call_later(MS, self.receive_eof_if_client)

        self.loop.run_until_complete(self.protocol.close(reason="close"))

        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(recv)

        self.assertConnectionClosed(CloseCode.NORMAL_CLOSURE, "close")

    # There is no test_remote_close_during_recv because it would be identical
    # to test_remote_close.

    def test_remote_close_during_send(self):
        self.make_drain_slow()
        send = self.loop.create_task(self.protocol.send("hello"))

        self.receive_frame(self.close_frame)
        self.receive_eof()

        with self.assertRaises(ConnectionClosed):
            self.loop.run_until_complete(send)

        self.assertConnectionClosed(CloseCode.NORMAL_CLOSURE, "close")

    # There is no test_local_close_during_send because this cannot really
    # happen, considering that writes are serialized.

    def test_broadcast_text(self):
        broadcast([self.protocol], "café")
        self.assertOneFrameSent(True, OP_TEXT, "café".encode("utf-8"))

    def test_broadcast_binary(self):
        broadcast([self.protocol], b"tea")
        self.assertOneFrameSent(True, OP_BINARY, b"tea")

    def test_broadcast_type_error(self):
        with self.assertRaises(TypeError):
            broadcast([self.protocol], ["ca", "fé"])

    def test_broadcast_no_clients(self):
        broadcast([], "café")
        self.assertNoFrameSent()

    def test_broadcast_two_clients(self):
        broadcast([self.protocol, self.protocol], "café")
        self.assertFramesSent(
            (True, OP_TEXT, "café".encode("utf-8")),
            (True, OP_TEXT, "café".encode("utf-8")),
        )

    def test_broadcast_skips_closed_connection(self):
        self.close_connection()

        with self.assertNoLogs():
            broadcast([self.protocol], "café")
        self.assertNoFrameSent()

    def test_broadcast_skips_closing_connection(self):
        close_task = self.half_close_connection_local()

        with self.assertNoLogs():
            broadcast([self.protocol], "café")
        self.assertNoFrameSent()

        self.loop.run_until_complete(close_task)  # cleanup

    def test_broadcast_skips_connection_sending_fragmented_text(self):
        self.make_drain_slow()
        self.loop.create_task(self.protocol.send(["ca", "fé"]))
        self.run_loop_once()
        self.assertOneFrameSent(False, OP_TEXT, "ca".encode("utf-8"))

        with self.assertLogs("websockets", logging.WARNING) as logs:
            broadcast([self.protocol], "café")

        self.assertEqual(
            [record.getMessage() for record in logs.records][:2],
            ["skipped broadcast: sending a fragmented message"],
        )

    @unittest.skipIf(
        sys.version_info[:2] < (3, 11), "raise_exceptions requires Python 3.11+"
    )
    def test_broadcast_reports_connection_sending_fragmented_text(self):
        self.make_drain_slow()
        self.loop.create_task(self.protocol.send(["ca", "fé"]))
        self.run_loop_once()
        self.assertOneFrameSent(False, OP_TEXT, "ca".encode("utf-8"))

        with self.assertRaises(ExceptionGroup) as raised:
            broadcast([self.protocol], "café", raise_exceptions=True)

        self.assertEqual(str(raised.exception), "skipped broadcast (1 sub-exception)")
        self.assertEqual(
            str(raised.exception.exceptions[0]), "sending a fragmented message"
        )

    def test_broadcast_skips_connection_failing_to_send(self):
        # Configure mock to raise an exception when writing to the network.
        self.protocol.transport.write.side_effect = RuntimeError

        with self.assertLogs("websockets", logging.WARNING) as logs:
            broadcast([self.protocol], "café")

        self.assertEqual(
            [record.getMessage() for record in logs.records][:2],
            ["skipped broadcast: failed to write message"],
        )

    @unittest.skipIf(
        sys.version_info[:2] < (3, 11), "raise_exceptions requires Python 3.11+"
    )
    def test_broadcast_reports_connection_failing_to_send(self):
        # Configure mock to raise an exception when writing to the network.
        self.protocol.transport.write.side_effect = RuntimeError("BOOM")

        with self.assertRaises(ExceptionGroup) as raised:
            broadcast([self.protocol], "café", raise_exceptions=True)

        self.assertEqual(str(raised.exception), "skipped broadcast (1 sub-exception)")
        self.assertEqual(str(raised.exception.exceptions[0]), "failed to write message")
        self.assertEqual(str(raised.exception.exceptions[0].__cause__), "BOOM")


class ServerTests(CommonTests, AsyncioTestCase):
    def setUp(self):
        super().setUp()
        self.protocol.is_client = False
        self.protocol.side = "server"

    def test_local_close_send_close_frame_timeout(self):
        self.protocol.close_timeout = 10 * MS
        self.make_drain_slow(50 * MS)
        # If we can't send a close frame, time out in 10ms.
        # Check the timing within -1/+9ms for robustness.
        with self.assertCompletesWithin(9 * MS, 19 * MS):
            self.loop.run_until_complete(self.protocol.close(reason="close"))
        self.assertConnectionClosed(CloseCode.ABNORMAL_CLOSURE, "")

    def test_local_close_receive_close_frame_timeout(self):
        self.protocol.close_timeout = 10 * MS
        # If the client doesn't send a close frame, time out in 10ms.
        # Check the timing within -1/+9ms for robustness.
        with self.assertCompletesWithin(9 * MS, 19 * MS):
            self.loop.run_until_complete(self.protocol.close(reason="close"))
        self.assertConnectionClosed(CloseCode.ABNORMAL_CLOSURE, "")

    def test_local_close_connection_lost_timeout_after_write_eof(self):
        self.protocol.close_timeout = 10 * MS
        # If the client doesn't close its side of the TCP connection after we
        # half-close our side with write_eof(), time out in 10ms.
        # Check the timing within -1/+9ms for robustness.
        with self.assertCompletesWithin(9 * MS, 19 * MS):
            # HACK: disable write_eof => other end drops connection emulation.
            self.transport._eof = True
            self.receive_frame(self.close_frame)
            self.run_loop_once()
            self.loop.run_until_complete(self.protocol.close(reason="close"))
        # Due to a bug in coverage, this is erroneously reported as not covered.
        self.assertConnectionClosed(  # pragma: no cover
            CloseCode.NORMAL_CLOSURE,
            "close",
        )

    def test_local_close_connection_lost_timeout_after_close(self):
        self.protocol.close_timeout = 10 * MS
        # If the client doesn't close its side of the TCP connection after we
        # half-close our side with write_eof() and close it with close(), time
        # out in 20ms.
        # Check the timing within -1/+9ms for robustness.
        # Add another 10ms because this test is flaky and I don't understand.
        with self.assertCompletesWithin(19 * MS, 39 * MS):
            # HACK: disable write_eof => other end drops connection emulation.
            self.transport._eof = True
            # HACK: disable close => other end drops connection emulation.
            self.transport._closing = True
            self.receive_frame(self.close_frame)
            self.run_loop_once()
            self.loop.run_until_complete(self.protocol.close(reason="close"))
        # Due to a bug in coverage, this is erroneously reported as not covered.
        self.assertConnectionClosed(  # pragma: no cover
            CloseCode.NORMAL_CLOSURE,
            "close",
        )


class ClientTests(CommonTests, AsyncioTestCase):
    def setUp(self):
        super().setUp()
        self.protocol.is_client = True
        self.protocol.side = "client"

    def test_local_close_send_close_frame_timeout(self):
        self.protocol.close_timeout = 10 * MS
        self.make_drain_slow(50 * MS)
        # If we can't send a close frame, time out in 20ms.
        # - 10ms waiting for sending a close frame
        # - 10ms waiting for receiving a half-close
        # Check the timing within -1/+9ms for robustness.
        with self.assertCompletesWithin(19 * MS, 29 * MS):
            self.loop.run_until_complete(self.protocol.close(reason="close"))
        # Due to a bug in coverage, this is erroneously reported as not covered.
        self.assertConnectionClosed(  # pragma: no cover
            CloseCode.ABNORMAL_CLOSURE,
            "",
        )

    def test_local_close_receive_close_frame_timeout(self):
        self.protocol.close_timeout = 10 * MS
        # If the server doesn't send a close frame, time out in 20ms:
        # - 10ms waiting for receiving a close frame
        # - 10ms waiting for receiving a half-close
        # Check the timing within -1/+9ms for robustness.
        with self.assertCompletesWithin(19 * MS, 29 * MS):
            self.loop.run_until_complete(self.protocol.close(reason="close"))
        # Due to a bug in coverage, this is erroneously reported as not covered.
        self.assertConnectionClosed(  # pragma: no cover
            CloseCode.ABNORMAL_CLOSURE,
            "",
        )

    def test_local_close_connection_lost_timeout_after_write_eof(self):
        self.protocol.close_timeout = 10 * MS
        # If the server doesn't half-close its side of the TCP connection
        # after we send a close frame, time out in 20ms:
        # - 10ms waiting for receiving a half-close
        # - 10ms waiting for receiving a close after write_eof
        # Check the timing within -1/+9ms for robustness.
        with self.assertCompletesWithin(19 * MS, 29 * MS):
            # HACK: disable write_eof => other end drops connection emulation.
            self.transport._eof = True
            self.receive_frame(self.close_frame)
            self.run_loop_once()
            self.loop.run_until_complete(self.protocol.close(reason="close"))
        # Due to a bug in coverage, this is erroneously reported as not covered.
        self.assertConnectionClosed(  # pragma: no cover
            CloseCode.NORMAL_CLOSURE,
            "close",
        )

    def test_local_close_connection_lost_timeout_after_close(self):
        self.protocol.close_timeout = 10 * MS
        # If the client doesn't close its side of the TCP connection after we
        # half-close our side with write_eof() and close it with close(), time
        # out in 30ms.
        # - 10ms waiting for receiving a half-close
        # - 10ms waiting for receiving a close after write_eof
        # - 10ms waiting for receiving a close after close
        # Check the timing within -1/+9ms for robustness.
        # Add another 10ms because this test is flaky and I don't understand.
        with self.assertCompletesWithin(29 * MS, 49 * MS):
            # HACK: disable write_eof => other end drops connection emulation.
            self.transport._eof = True
            # HACK: disable close => other end drops connection emulation.
            self.transport._closing = True
            self.receive_frame(self.close_frame)
            self.run_loop_once()
            self.loop.run_until_complete(self.protocol.close(reason="close"))
        # Due to a bug in coverage, this is erroneously reported as not covered.
        self.assertConnectionClosed(  # pragma: no cover
            CloseCode.NORMAL_CLOSURE,
            "close",
        )
